/*    $ JBOSN: drv_lan7110.c v 0.1 11-12-03 21:47 $    */

/*
 * Module Name: drv_lan7110.c
 *
 * Purpose: HMS30C7110 network device driver .
 *
 */

#include <types.h>
#include <thread.h>
#include <message.h>
#include <interrupt.h>
#include <device.h>
#include <error.h>
#include <network.h>
#include <memory.h>

#include <hal_panic.h>

#include <lib/string.h>

#include <mach/hal_interrupt.h>

#include <mach/SAMA5D3X.h>
#include <mach/SAMA5D3Xbd.h>

#include <mach/gpio.h>

#include "phy.h"
#include "drv_glan.h"


#define LAN_DBG_LVL   0

#define GMACB_PHY_RETRY_MAX            300000

static GmacStats gGmacStats;

/*----------------------------------------------------------------------------
 *        Macro
 *----------------------------------------------------------------------------*/
/** Return count in buffer */
#define GCIRC_CNT(head,tail,size) (((head) - (tail)) & ((size)-1))

/** Return space available, 0..size-1. always leave one free char as a completely full buffer 
    has head == tail, which is the same as empty */
#define GCIRC_SPACE(head,tail,size) GCIRC_CNT((tail),((head)+1),(size))

/** Return count up to the end of the buffer. Carefully avoid accessing head and tail more than once,
    so they can change underneath us without returning inconsistent results */
#define GCIRC_CNT_TO_END(head,tail,size) \
     ({int end = (size) - (tail); \
     int n = ((head) + end) & ((size)-1); \
     n < end ? n : end;})

/** Return space available up to the end of the buffer */
#define GCIRC_SPACE_TO_END(head,tail,size) \
   ({int end = (size) - 1 - (head); \
     int n = (end + (tail)) & ((size)-1); \
     n <= end ? n : end+1;})

/** Increment head or tail */
#define GCIRC_INC(headortail,size)      headortail = (((headortail) + 1) & (size - 1))


/** Circular buffer is empty ? */
#define GCIRC_EMPTY(head, tail)     (head == tail)

/** Clear circular buffer */
#define GCIRC_CLEAR(head, tail)  (head = tail = 0)

//------------------------------------------------------------------------------
//         Definitions
//------------------------------------------------------------------------------
/// The buffer addresses written into the descriptors must be aligned so the
/// last few bits are zero.  These bits have special meaning for the GMAC
/// peripheral and cannot be used as part of the address.
#define GMAC_ADDRESS_MASK   ((unsigned int)0xFFFFFFFC)
#define GMAC_LENGTH_FRAME   ((unsigned int)0x3FFF)    /// Length of frame mask

// receive buffer descriptor bits
#define GMAC_RX_OWNERSHIP_BIT   (1 <<  0)
#define GMAC_RX_WRAP_BIT        (1 <<  1)
#define GMAC_RX_SOF_BIT         (1 << 14)
#define GMAC_RX_EOF_BIT         (1 << 15)

// Transmit buffer descriptor bits
#define GMAC_TX_LAST_BUFFER_BIT (1 << 15)
#define GMAC_TX_WRAP_BIT        (1 << 30)
#define GMAC_TX_USED_BIT        (1 << 31)
#define GMAC_TX_RLE_BIT         (1 << 29) /// Retry Limit Exceeded
#define GMAC_TX_UND_BIT         (1 << 28) /// Tx Buffer Underrun
#define GMAC_TX_ERR_BIT         (1 << 27) /// Exhausted in mid-frame
#define GMAC_TX_ERR_BITS        (GMAC_TX_RLE_BIT | GMAC_TX_UND_BIT | GMAC_TX_ERR_BIT)



STATIC const struct pio_desc hw_pio_mode[] = {

	{"MODE0",	AT91C_PIN_PB(4), 0, PIO_DEFAULT, PIO_OUTPUT_1},
	{"MODE1",	AT91C_PIN_PB(5), 0, PIO_DEFAULT, PIO_OUTPUT_1},
	{"MODE2",	AT91C_PIN_PB(6), 0, PIO_DEFAULT, PIO_OUTPUT_1},
	{"MODE3",	AT91C_PIN_PB(7), 0, PIO_DEFAULT, PIO_OUTPUT_1},	
};

STATIC const struct pio_desc hw_pio_run[] = {
	{"GTX0",	AT91C_PIN_PB( 0), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GTX1",	AT91C_PIN_PB( 1), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GTX2",	AT91C_PIN_PB( 2), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GTX3",	AT91C_PIN_PB( 3), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRX0",	AT91C_PIN_PB( 4), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRX1",	AT91C_PIN_PB( 5), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRX2",	AT91C_PIN_PB( 6), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRX3",	AT91C_PIN_PB( 7), 0, PIO_DEFAULT, PIO_PERIPH_A},	
	{"GTXCK",	AT91C_PIN_PB( 8), 0, PIO_DEFAULT, PIO_PERIPH_A},						 
	{"GTXEN",	AT91C_PIN_PB( 9), 0, PIO_DEFAULT, PIO_PERIPH_A},
	//{"ETXER",	AT91C_PIN_PB(10), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"ERXCK",	AT91C_PIN_PB(11), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRXDV",	AT91C_PIN_PB(12), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GRXER",	AT91C_PIN_PB(13), 0, PIO_DEFAULT, PIO_PERIPH_A},
	//{"GCRS",	AT91C_PIN_PB(14), 0, PIO_DEFAULT, PIO_PERIPH_A},
	//{"GCOL",	AT91C_PIN_PB(15), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GMDC",	AT91C_PIN_PB(16), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"GMIDO",	AT91C_PIN_PB(17), 0, PIO_DEFAULT, PIO_PERIPH_A},
	{"G125CK",	AT91C_PIN_PB(18), 0, PIO_DEFAULT, PIO_PERIPH_A},

	{"GINT",	AT91C_PIN_PB(25), 0, PIO_PULLUP,  PIO_INPUT},
};

//__attribute__((aligned(64)))
//static GMACRXDESC RxDs[RX_BUFFERS];
//
//__attribute__((aligned(64)))
//static GMACTXDESC TxDs[TX_BUFFERS];
//
//__attribute__((aligned(64))) 
//STATIC UINT32   TxBuffer[TX_BUFFERS * GMAC_TX_UNITSIZE];
//
//__attribute__((aligned(64))) 
//STATIC UINT32   TxBuffer[RX_BUFFERS * GMAC_RX_UNITSIZE];
//
//
//
//#define DEF_TX_RING_PENDING    (TX_RING_SIZE - 1)

/* Make the IP header word-aligned (the ethernet header is 14 bytes) */
#define RX_OFFSET        2

//#define TX_RING_GAP(pEthNetInfo)        (TX_RING_SIZE - (pEthNetInfo)->TxPending)
//
//#define TX_BUFFS_AVAIL(pEthNetInfo)     (((pEthNetInfo)->TxTail <= (pEthNetInfo)->TxHead) ? \
//                                        (pEthNetInfo)->TxTail + DEF_TX_RING_PENDING - (pEthNetInfo)->TxHead :    \
//                                        (pEthNetInfo)->TxTail - (pEthNetInfo)->TxHead - 1)
//
//#define NEXT_TX(n)                        (((n) + 1) & (TX_RING_SIZE - 1))
//#define NEXT_RX(n)                        (((n) + 1) & (RX_RING_SIZE - 1))
//
///* minimum number of free TX descriptors before waking up TX process */
//#define MACB_TX_WAKEUP_THRESH    (TX_RING_SIZE / 4)
//
//#define MACB_RX_INT_FLAGS    (MACB_BIT(RCOMP) | MACB_BIT(RXUBR) | MACB_BIT(ISR_ROVR))

#define DMA_ADDR_OFFSET         0x20000000

NET_DEV LAN_NetDev =
{
        NULL,
        "SAMA5D3X GNET Interface0",
        DEV_TYPE_ETHER,

        (192<<0)+(168<<8)+(11<<16)+(30<<24),            // 192.168.11.30
        (255<<0)+(255<<8)+(255<<16)+(128<<24),          // 255.255.255.128
        (192<<0)+(168<<8)+(11<<16)+(1<<24),              // 192.168.11.1
};

// net configuration
#define ETH_ALEN                6

STATIC UINT32       RxPackets = 0;
STATIC UINT32       RxBytes = 0;

STATIC UINT32       TxLen = 0;

//STATIC INT32        iNetIf;

// for driver thread
#define msWait(x)               usWait((x)*2*1000)
#define LANDEV_STACK_SIZE       (4096)

STATIC UINT32                   LANDev_Stack[LANDEV_STACK_SIZE>>2];
STATIC BOOL                     thLANDeviceDriver( UINT32 Arg );
STATIC DEVICECONFIG             LAN_DeviceConfig = {"ETH", 0, -1, NULL, NULL};

PUBLIC THREADCONFIG             GMACB_LANDev_Thread = 
                                {
                                    (UINT32)&LAN_NetDev, 
                                    thLANDeviceDriver, 
                                    LANDev_Stack, 
                                    LANDEV_STACK_SIZE, 
                                    THREAD_PRIORITY_DRIVER_HIGH+3
                                };

STATIC ENET_INFO        EthNetInfo;
//STATIC ENET_BUFDESC     TxBufDesc[TX_BD_DEFAULT];
//STATIC ENET_BUFDESC     RxBufDesc[RX_BD_DEFAULT];


// for LAN
#define min(a,b)    ((a>b)?b:a)
#define max(a,b)    ((a<b)?b:a)

STATIC AT91PS_GMAC      pGMAC = (AT91PS_GMAC)AT91C_BASE_GMAC;

EXTERN PHY_INTERFACE DM9161A_PhyInterface;
EXTERN PHY_INTERFACE KSZ9201_PhyInterface;


//#define AT91C_EMAC_PHY_SOF (0x01 << 30)
//#define AT91C_EMAC_PHY_CODE (0x02 << 16)
//#define AT91C_EMAC_PHY_BASE_WORD (AT91C_EMAC_PHY_SOF | AT91C_EMAC_PHY_CODE)
//
//#define AT91C_EMAC_READ_PHY (0x02 << 28)
//#define AT91C_EMAC_WRITE_PHY (0x01 << 28)

/*
 * Wait PHY operation complete.
 * Return 1 if the operation completed successfully.
 */
STATIC BOOL
WaitPhy( AT91PS_GMAC pGMAC, UINT32 Retry )
{
    VOLATILE UINT32 nRetry = 0;
    
    while( (pGMAC->GMAC_NSR & GMAC_NSR_IDLE) == 0 )
    {
        if(Retry == 0)
            continue;
        nRetry ++;
        if (nRetry >= Retry)
        {
            return 0;
        }
    }
    return 1;
}

/*
 * Read value stored in a PHY register.
 * Note: MDI interface is assumed to already have been enabled.
 */
STATIC VOID
ReadPhy( AT91PS_GMAC pGMAC, unsigned char phy_addr, unsigned char address, DWORD *dwValue)
{
    /* Wait until bus idle */
    WaitPhy(pGMAC, 0);

    /* Write register */
    pGMAC->GMAC_MAN = (~GMAC_MAN_WZO & GMAC_MAN_CLTTO)
                      | (GMAC_MAN_OP(1 ? 0x2 : 0x1))
                      | GMAC_MAN_WTN(0x02)
                      | GMAC_MAN_PHYA(phy_addr)
                      | GMAC_MAN_REGA(address)
                      | GMAC_MAN_DATA(0) ;

    /* Wait until bus idle */
    WaitPhy(pGMAC, 0);

    /* read value */
    *dwValue = pGMAC->GMAC_MAN & GMAC_MAN_DATA_Msk;

    return;
}


/*
 * Write value stored in a PHY register.
 * Note: MDI interface is assumed to already have been enabled.
 */
STATIC BOOL
WritePhy(AT91PS_GMAC pGMAC, unsigned char phy_addr, unsigned char address, DWORD dwValue)
{
    /* Wait until bus idle */
    WaitPhy(pGMAC, 0);

    /* Write register */
    pGMAC->GMAC_MAN = (~GMAC_MAN_WZO & GMAC_MAN_CLTTO)
                      | (GMAC_MAN_OP(0 ? 0x2 : 0x1))
                      | GMAC_MAN_WTN(0x02)
                      | GMAC_MAN_PHYA(phy_addr)
                      | GMAC_MAN_REGA(address)
                      | GMAC_MAN_DATA(dwValue) ;


	/* Wait until IDLE bit in Network Status register is cleared */
    if ( WaitPhy(pGMAC, GMACB_PHY_RETRY_MAX) == 0 ) 
        return FALSE;

    return TRUE;
}

STATIC VOID
AT91F_Enable_Mdi(AT91PS_GMAC pGMAC)
{
    pGMAC->GMAC_NCR &=  ~(GMAC_NCR_RXEN | GMAC_NCR_TXEN);
    pGMAC->GMAC_NCR |= GMAC_NCR_MPE;
    pGMAC->GMAC_NCR |=  (GMAC_NCR_RXEN | GMAC_NCR_TXEN);
	//pGMAC->EMAC_NCR |= AT91C_EMAC_MPE;	/* enable management port */
}

/*
 * Disable the MDIO bit in the MAC control register
 */
STATIC VOID
AT91F_Disable_Mdi(AT91PS_GMAC pGMAC)
{
    pGMAC->GMAC_NCR &=  ~(GMAC_NCR_RXEN | GMAC_NCR_TXEN);
    pGMAC->GMAC_NCR &= ~GMAC_NCR_MPE;
    pGMAC->GMAC_NCR |=  (GMAC_NCR_RXEN | GMAC_NCR_TXEN);
	//pGMAC->EMAC_NCR &= ~AT91C_EMAC_MPE;	/* disable management port */
}



//STATIC VOID
//PhyRegEnable( PENET_INFO pEthNetInfo )
//{
//    //UINT32  Flags;
//    UINT32  Reg;
//    
//    Reg = pGMAC->EMAC_NCR;
//    Reg |= MACB_BIT(MPE);
//    pGMAC->EMAC_NCR = Reg;
//    pGMAC->EMAC_IER = MACB_BIT(MFD);
//
//    return;
//}
//
//STATIC VOID
//PhyRegDisable( PENET_INFO pEthNetInfo )
//{
//    //UINT32  Flags;
//    UINT32  Reg;
//
//    Reg = pGMAC->EMAC_NCR;
//    Reg &= ~MACB_BIT(MPE);
//    pGMAC->EMAC_NCR = Reg;
//    pGMAC->EMAC_IDR = MACB_BIT(MFD);
//
//    return;
//}
//
//
//STATIC UINT32
//PhyRegRead( PENET_INFO pEthNetInfo, UINT32 PhyAddr, UINT32 nReg )
//{
//    VOLATILE UINT32      Temp;
//
//    PhyRegEnable( pEthNetInfo );
//    
//    pGMAC->EMAC_MAN =   MACB_BF(SOF, MACB_MAN_SOF) |
//                        MACB_BF(RW, MACB_MAN_READ) |
//                        MACB_BF(PHYA, PhyAddr) |
//                        MACB_BF(REGA, nReg) |
//                        MACB_BF(CODE, MACB_MAN_CODE);
//
//    //WaitForCompletion();
//{
//    VOLATILE UINT32  Status;
//    Sleep(1);
//    do
//    {
//        Status = pGMAC->EMAC_ISR;
//    }while( !(Status & MACB_BIT(MFD)) );
//
//}
//
//    Temp = MACB_BFEXT(DATA, pGMAC->EMAC_MAN);
//
//    PhyRegDisable( pEthNetInfo );
//
//
//    return Temp;
//
////    // set read command
////    MAC_REG_WRITE( pEthNetInfo, MII_CMD, 0x40000 | (pEthNetInfo->PhyAddr << 8) | nReg );
////
////    // wait until complete read
////    do
////    {
////        MAC_REG_READ(pEthNetInfo, MII_CMD, Temp );
////    }while( Temp & 0x40000);
////    
////    MAC_REG_READ( pEthNetInfo, MII_RXDATA, Temp );
////    
////    return Temp & 0xFFFF;
//}
//
//
//
//STATIC VOID
//PhyRegWrite( PENET_INFO pEthNetInfo, UINT32 PhyAddr, UINT32 nReg, UINT32 Value )
//{
//
//    PhyRegEnable(pEthNetInfo);
//    
//    pGMAC->EMAC_MAN =   MACB_BF(SOF, MACB_MAN_SOF) |
//                        MACB_BF(RW, MACB_MAN_WRITE) |
//                        MACB_BF(PHYA, PhyAddr) |
//                        MACB_BF(REGA, nReg) |
//                        MACB_BF(CODE, MACB_MAN_CODE)|
//                        MACB_BF(DATA, Value);
//
//    //WaitForCompletion();
//{
//    VOLATILE UINT32  Status;
//    do
//    {
//        Sleep(1);
//        Status = pGMAC->EMAC_ISR;
//    }while( !(Status & MACB_BIT(MFD)) );
//
//}
//
//    PhyRegDisable(pEthNetInfo);
//
//    return;
//
////    VOLATILE UINT32      Temp;
////
////    MAC_REG( pEthNetInfo, MII_TXDATA ) = Value;
////    MAC_REG( pEthNetInfo, MII_CMD ) = 0x20000 | (pEthNetInfo->PhyAddr << 8) | nReg;
////
////    // wait until complete writing
////    do
////    {
////        Temp  = MAC_REG( pEthNetInfo, MII_CMD );
////    }while( Temp & 0x20000); 
////
////    return;
//}
//
//STATIC UINT32
//PhyLinkModeRead( PENET_INFO pEthNetInfo )
//{
//    UINT32      RetVal;
//
//    RetVal = MAC_MODE_100MBPS | MAC_MODE_FULLDUPLEX;
//    
//    return RetVal;
//}


STATIC BOOL
PhyLinkStatusRead( PENET_INFO pEthNetInfo )
{
    UINT32  PhyChipStatus;
    
    // read chip status
    //PhyChipStatus = PhyRegRead( pEthNetInfo, PHY_STATUS );

    return TRUE;

    if( PhyChipStatus & 0x4 )
        return TRUE;            // PHY link-up
    else
        return FALSE;           // PHY link-down
}



//STATIC VOID
//MACB_SetMedia( PENET_INFO pEthNetInfo, UINT32 Mode )
//{
//    UINT32      Reg;
//
//    Reg = pGMAC->EMAC_NCFGR;
//    Reg &= ~(MACB_BIT(SPD) | MACB_BIT(FD));
//    
//    if( Mode & MAC_MODE_100MBPS )
//        Reg |= MACB_BIT(SPD);
//    
//    if( Mode & MAC_MODE_FULLDUPLEX )
//        Reg |= MACB_BIT(FD);
//
//    pGMAC->EMAC_NCFGR = Reg;
//
//    return;
//}
//
//
//STATIC BOOL
//PhyProbe( PENET_INFO pEthNetInfo )
//{
//    UINT32      PhyAddr;
//    UINT16      PhyId1, PhyId2;
//    UINT32      PhyId;
//
//
//
//    
//    for( PhyAddr = 0; PhyAddr < 32 ; PhyAddr++)
//    {
//        PhyId1 = PhyRegRead( pEthNetInfo, PhyAddr, MII_PHYSID1 );
//        PhyId2 = PhyRegRead( pEthNetInfo, PhyAddr, MII_PHYSID2 );
//
////DEBUG(LAN_DBG_LVL,"ID = (0x%x, 0x%x)\n", PhyId1, PhyId2);
//
//#if 0
//        PhyId = (PhyId1 << 16) | (PhyId2 & 0xfff0);
//        switch (PhyId)
//        {
//            case MII_DM9161_ID:         /* Davicom 9161: PHY_ID1 = 0x181, PHY_ID2 = B881 */
//            case MII_DM9161A_ID:        /* Davicom 9161A: PHY_ID1 = 0x181, PHY_ID2 = B8A0 */
//            case MII_LXT971A_ID:        /* Intel LXT971A: PHY_ID1 = 0x13, PHY_ID2 = 78E0 */
//            case MII_RTL8201_ID:        /* Realtek RTL8201: PHY_ID1 = 0, PHY_ID2 = 0x8201 */
//            case MII_BCM5221_ID:        /* Broadcom BCM5221: PHY_ID1 = 0x40, PHY_ID2 = 0x61e0 */
//            case MII_DP83847_ID:        /* National Semiconductor DP83847:  */
//            case MII_DP83848_ID:        /* National Semiconductor DP83848:  */
//            case MII_AC101L_ID:         /* Altima AC101L: PHY_ID1 = 0x22, PHY_ID2 = 0x5520 */
//            case MII_KS8721_ID:         /* Micrel KS8721: PHY_ID1 = 0x22, PHY_ID2 = 0x1610 */
//            case MII_T78Q21x3_ID:       /* Teridian 78Q21x3: PHY_ID1 = 0x0E, PHY_ID2 = 7237 */
//            case MII_LAN83C185_ID:      /* SMSC LAN83C185: PHY_ID1 = 0x0007, PHY_ID2 = 0xC0A1 */
//                detected = at91ether_setup(phy_id, phy_address, pdev, ether_clk);
//                break;
//        }
//#else
//        PhyId = (PhyId1 << 16) | (PhyId2 & 0xfff0);
//        if( PhyId != 0xFFFF0000 && PhyId != 0xFFFFFFF0 &&
//            PhyId != 0x0000FFF0 && PhyId != 0x00000000 )
//            break;
//#endif
//
//    }
//    if( PhyAddr == 32 )
//    {
//        DEBUG1(LAN_DBG_LVL," NO PHY CHIP found\n");
//        return FALSE;
//    }
//
//    kprintf("PHY CHIP found : addr = %x, ID(%04x, %04x)\n", PhyAddr, PhyId1, PhyId2);
//
//    pEthNetInfo->PhyAddr = PhyAddr;
//    pEthNetInfo->PhyId = PhyId;
//
//
//
//    PhyRegWrite( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR, (0x1<<15) |PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR ));    // reset
//
//    Sleep(1);
//
//
//    return TRUE;
//}
//
//
//STATIC VOID
//PhySetup( PENET_INFO pEthNetInfo )
//{
//    UINT32  Mode;
//    UINT32  Ctrl;
//
//    PhyProbe( pEthNetInfo );
//
//    Ctrl = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR );
//    Mode = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMSR );
//DEBUGMSG(LAN_DBG_LVL,"PhySetup : Ctrl = %x\n", Ctrl);
//DEBUGMSG(LAN_DBG_LVL,"PhySetup : Mode = %x\n", Mode);
//
////    Mode = 0 ;
//    Mode = MAC_MODE_100MBPS | MAC_MODE_FULLDUPLEX;
////    Mode = MAC_MODE_100MBPS ;
//
//    Ctrl = 0;
//    if( Mode & MAC_MODE_100MBPS )
//        Ctrl |= BMCR_SPEED100;
//    if( Mode & MAC_MODE_FULLDUPLEX )
//        Ctrl |= BMCR_FULLDPLX;
//
//    PhyRegWrite( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR, Ctrl );
//
//    Ctrl = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMCR );
//    Mode = PhyRegRead( pEthNetInfo, pEthNetInfo->PhyAddr, MII_BMSR );
//DEBUGMSG(LAN_DBG_LVL,"PhySetup : Ctrl = %x\n", Ctrl);
//DEBUGMSG(LAN_DBG_LVL,"PhySetup : Mode = %x\n", Mode);
//
//    MACB_SetMedia( pEthNetInfo, Mode );
//
//    //PhyRegWrite( pEthNetInfo, MII_ADVERTISE, ADVERTISE_CSMA | ADVERTISE_ALL );
//    //PhyRegWrite( pEthNetInfo, MII_BMCR, (BMCR_SPEED100 | BMCR_ANENABLE | BMCR_ANRESTART | BMCR_FULLDPLX));
//
//}

STATIC VOID
PhySetup( PENET_INFO pEthNetInfo, BOOL bFullDuplex, UINT32 Speed )
{
//void GMAC_EnableRGMII(Gmac *pGMAC, BOOL bFullDuplex, UINT32 Speed)
//{
    pGMAC->GMAC_NCR &=  ~(GMAC_NCR_RXEN | GMAC_NCR_TXEN);
    pGMAC->GMAC_NCFGR &= ~GMAC_NCFGR_GBE;
    if( bFullDuplex == FALSE )
    {
        pGMAC->GMAC_NCFGR &= ~GMAC_NCFGR_FD;
    }
    else 
    {
        pGMAC->GMAC_NCFGR |= GMAC_NCFGR_FD;
    }


    if (Speed == SPEED_10)
    {
        pGMAC->GMAC_NCFGR &= ~GMAC_NCFGR_SPD;
    }
    else
    if(Speed == SPEED_100) 
    {
        pGMAC->GMAC_NCFGR |= GMAC_NCFGR_SPD;
    }
    else
    {
        pGMAC->GMAC_NCFGR |= GMAC_NCFGR_GBE;
    }
  
    /* RGMII enable */
    pGMAC->GMAC_UR |= GMAC_UR_RGMII;



    return;
}


//*----------------------------------------------------------------------------
//* \fn    PhyInit
//* \brief This function initialise the ethernet
//*----------------------------------------------------------------------------
STATIC BOOL
PhyInit( PENET_INFO pEthNetInfo )
{
	UINT32              MacConfig;
	MAC_PHY_OP			MacPhyOp;
	PHY_INFO            PhyInfo;
	PHY_PINTERFACE      pPhyInterface;
	BYTE                PhyAddr;	
	PHY_PINTERFACE      PhyList[] = 
	{		
		//&DM9161A_PhyInterface,
		&KSZ9201_PhyInterface,
	};


	MacPhyOp.ReadPhy = ReadPhy;
	MacPhyOp.WritePhy = WritePhy;
	MacPhyOp.pDeviceLocation = pGMAC;

	DEBUGMSG(1,"->Enter in PhyInit \n\r");



	/* Read the PHY ID registers */
	AT91F_Enable_Mdi(pGMAC);

    /* reset PHY if you need */
	
	pPhyInterface = FindPhy(&MacPhyOp, PhyList, sizeof(PhyList)/sizeof(PhyList[0]), &PhyAddr);
	if( pPhyInterface == NULL )
	{
		DEBUGMSG(1,"No Phy Found\r\n");
		return -1;
	}
	DEBUGMSG(1,"Found Phy (%s) at address %d\r\n",pPhyInterface->pName,PhyAddr);

	//DEBUGMSG(1,"%s(%d): reset\r\n",__FUNCTION__,__LINE__);
    /* set mode */
	pio_setup(hw_pio_mode, sizeof(hw_pio_mode)/sizeof(struct pio_desc));

    Sleep(100);
    pPhyInterface->PHY_Reset(&MacPhyOp,PhyAddr);
    Sleep(100);

	pio_setup(hw_pio_run, sizeof(hw_pio_run)/sizeof(struct pio_desc));
    Sleep(100);

	//pPhyInterface->PHY_Reset(&MacPhyOp,PhyAddr);
#define GET_LINK_SPEED_TIMEOUT 20000
	
#if 0
//	PhyInfo.bFullDuplex = FALSE;
//	PhyInfo.bAutoNegociation = FALSE;
//	PhyInfo.Speed = SPEED_10;
//	pPhyInterface->PHY_SetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo);

	PhyInfo.bAutoNegociation = TRUE;
	pPhyInterface->PHY_SetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo);
	if (pPhyInterface->PHY_WaitForAutonegociationComplete(&MacPhyOp,PhyAddr,GET_LINK_SPEED_TIMEOUT) == FALSE)
	{
		DEBUGMSG(1,"Autonegociation failed\r\n");
		return -1;
	}
#else
	PhyInfo.bAutoNegociation = FALSE;
	if(pPhyInterface->PHY_GetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo) == FALSE)
	{
		if (PhyInfo.bAutoNegociation == TRUE) 
		{
			// Autonegociation is started but not complete yet
		}
		else
		{	 // Autonegocitaion is not selected
			DEBUGMSG(1,"GMAC: Starting Autonegociation\r\n");
			pPhyInterface->PHY_StartAutoNegociation(&MacPhyOp,PhyAddr);
		}
		if (pPhyInterface->PHY_WaitForAutonegociationComplete(&MacPhyOp,PhyAddr,GET_LINK_SPEED_TIMEOUT) == FALSE)
		{
			DEBUGMSG(1,"Autonegociation failed\r\n");
			return -1;
		}
	}
#endif


	if (pPhyInterface->PHY_WaitForLink(&MacPhyOp,PhyAddr,GET_LINK_SPEED_TIMEOUT) == FALSE)
	{
		DEBUGMSG(1,"No Link\r\n");
		return -1;
	}

	if(pPhyInterface->PHY_GetConfiguration(&MacPhyOp,PhyAddr,&PhyInfo) == FALSE)
	{
		DEBUGMSG(1,"Unable to get the network configuration from the PHY\r\n");
		return -1;
	
	}

	DEBUGMSG(1,"\n\r EMAC Init : %d Mbit/s %s DUPLEX (%s)\r\n", PhyInfo.Speed,PhyInfo.bFullDuplex ? L"FULL" : L"HALF", PhyInfo.bRMII ? L"RMII" : L"MII");


	/* Read the PHY ID registers */
	AT91F_Disable_Mdi(pGMAC);



//    PIO_Configure(pEmacPins, nbEmacPins);

    {
        UINT32      clock_dividor;

        pGMAC->GMAC_NCR &=  ~(GMAC_NCR_RXEN | GMAC_NCR_TXEN);
        if (BOARD_MCK <= 20000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_8;          // MDC clock = MCK/8
        else
        if (BOARD_MCK <= 40000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_16;         // MDC clock = MCK/16
        else
        if (BOARD_MCK <= 80000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_32;         // MDC clock = MCK/32
        else
        if (BOARD_MCK <= 160000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_64;         // MDC clock = MCK/64
        else
        if (BOARD_MCK <= 240000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_96;         // MDC clock = MCK/96
        else
        if (BOARD_MCK <= 320000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_128;        // MDC clock = MCK/128
        else
        if (BOARD_MCK <= 540000000)
            clock_dividor = GMAC_NCFGR_CLK_MCK_224;        // MDC clock = MCK/224
        else
            return 0;
    
        pGMAC->GMAC_NCFGR = (pGMAC->GMAC_NCFGR & (~GMAC_NCFGR_CLK_Msk)) | clock_dividor;
        pGMAC->GMAC_NCR |=  (GMAC_NCR_RXEN | GMAC_NCR_TXEN);
    }

//        /* Check PHY Address */
//        phy = GMACB_FindValidPhy(pMacb);
//        if (phy == 0xFF)
//        {
//            TRACE_ERROR("PHY Access fail\n\r");
//            return 0;
//        }
//        if(phy != pMacb->phyAddress) {
//            pMacb->phyAddress = phy;
//            GMACB_ResetPhy(pMacb);
//        }
//    }
//    else
//    {
//        TRACE_ERROR("PHY Reset Timeout\n\r");
//    }
//    return rc;
//}

    PhySetup(pEthNetInfo, PhyInfo.bFullDuplex, PhyInfo.Speed);

//
//	pGMAC->EMAC_NCR  = 0;
//
//	/* Update the MAC with the new information */
//	MacConfig = pGMAC->EMAC_NCFGR & ~(AT91C_EMAC_SPD | AT91C_EMAC_FD);
//	
//	if( PhyInfo.Speed == SPEED_100 )
//	{
//		MacConfig |= AT91C_EMAC_SPD;
//	}
//	else if (PhyInfo.Speed == SPEED_10) 
//	{
//		//Nothing to do here
//	}
//	else
//	{
//		DEBUGMSG(1,"Invalid network speed\r\n");
//	}
//
//	if (PhyInfo.bFullDuplex== TRUE)	
//	{
//		MacConfig |= AT91C_EMAC_FD;
//	}
//
//	
//
//	pGMAC->EMAC_NCFGR = MacConfig;
//	pGMAC->EMAC_RSR  = 0xFFFFFFFF;
//	pGMAC->EMAC_TSR  = 0xFFFFFFFF;
//	pGMAC->EMAC_NCR  |= AT91C_EMAC_CLRSTAT; // Clear the statistics
//	pGMAC->EMAC_USRIO  = (PhyInfo.bRMII ? AT91C_EMAC_RMII : 0) | AT91C_EMAC_CLKEN;
//	pGMAC->EMAC_NCR  |= (AT91C_EMAC_TE | AT91C_EMAC_RE);
//
//	//{ DWORD dummy = pGMAC->EMAC_ISR;}
//	//pGMAC->EMAC_IER = 0x0000080;
//	//{ DWORD dummy = pGMAC->EMAC_ISR;}


 	return 0;
}


//========================================

/**
 *  \brief Disable TX & reset registers and descriptor list
 *  \param pDrv Pointer to GMAC Driver instance.
 */
STATIC VOID 
MacResetTx( PENET_INFO pEthNetInfo )
{
    /* Disable TX */
    pGMAC->GMAC_NCR &= ~GMAC_NCR_TXEN;

    /* Transmit Buffer Queue Pointer Register */
    pGMAC->GMAC_TBQB = GMAC_TBQB_ADDR_Msk & (UINT32)pEthNetInfo->pTxTd;

    return;
}
    
/**
 *  \brief Disable RX & reset registers and descriptor list
 *  \param pDrv Pointer to GMAC Driver instance. 
 */
STATIC VOID 
MacResetRx( PENET_INFO pEthNetInfo )
{
    /* Disable RX */
    pGMAC->GMAC_NCR &= ~GMAC_NCR_RXEN;

    /* Receive Buffer Queue Pointer Register */
    pGMAC->GMAC_RBQB = GMAC_RBQB_ADDR_Msk & (UINT32)pEthNetInfo->pRxTd;

    return;
}

// reset the MAC
STATIC VOID
MacReset( PENET_INFO pEthNetInfo )
{
     /* Reset TX & RX */
    MacResetRx(pEthNetInfo);
    MacResetTx(pEthNetInfo);

//   UINT32      Temp;
//
//    /*
//     * Disable RX and TX (XXX: Should we halt the transmission
//     * more gracefully?)
//     */
//    pGMAC->EMAC_NCR = 0;
//    //macb_writel(bp, NCR, 0);
//
//
//    /* Clear the stats registers (XXX: Update stats first?) */
//    pGMAC->EMAC_NCR = MACB_BIT(CLRSTAT);
//    //macb_writel(bp, NCR, MACB_BIT(CLRSTAT));
//
//    /* Clear all status flags */
//    pGMAC->EMAC_TSR = ~0x0;
//    pGMAC->EMAC_RSR = ~0x0;
//    //macb_writel(bp, TSR, ~0UL);
//    //macb_writel(bp, RSR, ~0UL);
//
//    /* Disable all interrupts */
//    pGMAC->EMAC_IDR = ~0x0;
//    Temp = pGMAC->EMAC_ISR;
//    //macb_writel(bp, IDR, ~0UL);
//    //macb_readl(bp, ISR);
//

    return;
}


// initialize the MAC
STATIC VOID
MacInit( PENET_INFO pEthNetInfo, NET_PDEV pNetDev  )
{
    UINT32  Config;




//    // config the interface
//    Config = pGMAC->EMAC_NCFGR & MACB_BF(CLK, -1L);
//    Config |= MACB_BIT(PAE);        /* Pause Enable */
//    Config |= MACB_BIT(DRFCS);        /* Discard Rx FCS */
//
//    //if( pEthNetInfo->Dev->Flags & IFF_PROMISC)
//        Config |= MACB_BIT(CAF);    /* Copy All Frames */
//    //if(!(pEthNetInfo->Dev->Flags & IFF_BROADCAST))
//    //    Config |= MACB_BIT(NBC);    /* No BroadCast */
//
//    pGMAC->EMAC_NCFGR = Config;
//
//    /* Initialize TX and RX buffers */
//#if 0
//    pGMAC->EMAC_RBQP = ((UINT32)pEthNetInfo->pRxRing) - OFFSET_NONCACHE;  // physical address
//    pGMAC->EMAC_TBQP = ((UINT32)pEthNetInfo->pTxRing) - OFFSET_NONCACHE;  // physical address
//#else
//    pGMAC->EMAC_RBQP = ((UINT32)pEthNetInfo->pRxRing);  // physical address
//    pGMAC->EMAC_TBQP = ((UINT32)pEthNetInfo->pTxRing);  // physical address
//#endif
//
//    /* Assign RX buffers */
//    if (   ((uint32_t)pRxBuffer & 0x7)
//        || ((uint32_t)pRxD      & 0x7) )
//    {
//        wRxSize --;
//        TRACE_DEBUG("RX list address adjusted\n\r");
//    }
//    pGmacd->pRxBuffer = (uint8_t*)((uint32_t)pRxBuffer & 0xFFFFFFF8);
//    pGmacd->pRxD = (sGmacRxDescriptor*)((uint32_t)pRxD & 0xFFFFFFF8);
//    pGmacd->wRxListSize = wRxSize;
//
//    /* Assign TX buffers */
//    if (   ((uint32_t)pTxBuffer & 0x7)
//        || ((uint32_t)pTxD      & 0x7) )
//    {
//        wTxSize --;
//        TRACE_DEBUG("TX list address adjusted\n\r");
//    }
//    pGmacd->pTxBuffer = (uint8_t*)((uint32_t)pTxBuffer & 0xFFFFFFF8);
//    pGmacd->pTxD = (sGmacTxDescriptor*)((uint32_t)pTxD & 0xFFFFFFF8);
//    pGmacd->wTxListSize = wTxSize;
//    pGmacd->fTxCbList = pTxCb;


    /* Enable Rx and Tx, plus the stats register. */
    pGMAC->GMAC_NCR |=  GMAC_NCR_TXEN;
    pGMAC->GMAC_NCR |=  GMAC_NCR_RXEN;
    pGMAC->GMAC_NCR |=  GMAC_NCR_WESTAT;

    /* Enable interrupts */
    /* Setup the interrupts for TX (and errors) */
    pGMAC->GMAC_IER =  GMAC_IER_MFS 
                      |GMAC_IER_RCOMP
                      |GMAC_IER_RXUBR
                      |GMAC_IER_TXUBR
                      |GMAC_IER_TUR
                      |GMAC_IER_RLEX
                      |GMAC_IER_TFC 
                      |GMAC_IER_TCOMP
                      |GMAC_IER_ROVR 
                      |GMAC_IER_HRESP
                      |GMAC_IER_PFNZ 
                      |GMAC_IER_PTZ 
                      |GMAC_IER_PFTR
                      |GMAC_IER_EXINT
                      |GMAC_IER_DRQFR
                      |GMAC_IER_SFR
                      |GMAC_IER_DRQFT
                      |GMAC_IER_SFT
                      |GMAC_IER_PDRQFR
                      |GMAC_IER_PDRSFR
                      |GMAC_IER_PDRQFT
                      |GMAC_IER_PDRSFT;


    // set HW ETHERNET ADDRESS
    pGMAC->GMAC_SA[0].GMAC_SAB =   (pNetDev->EthAddr[3] << 24)
                                 | (pNetDev->EthAddr[2] << 16)
                                 | (pNetDev->EthAddr[1] <<  8)
                                 | (pNetDev->EthAddr[0]      )
                                 ;
    pGMAC->GMAC_SA[0].GMAC_SAT =   (pNetDev->EthAddr[5] <<  8)
                                 | (pNetDev->EthAddr[4]      )
                                 ;



    return;
}



// start the MAC
STATIC VOID
MacStart( PENET_INFO pEthNetInfo )
{

    /* Enable Rx and Tx, plus the stats register. */
    pGMAC->GMAC_NCR |=  GMAC_NCR_TXEN;
    pGMAC->GMAC_NCR |=  GMAC_NCR_RXEN;
    pGMAC->GMAC_NCR |=  GMAC_NCR_WESTAT;

    /* Enable interrupts */
    /* Setup the interrupts for TX (and errors) */
    pGMAC->GMAC_IER =  GMAC_IER_MFS 
                      |GMAC_IER_RCOMP
                      |GMAC_IER_RXUBR
                      |GMAC_IER_TXUBR
                      |GMAC_IER_TUR
                      |GMAC_IER_RLEX
                      |GMAC_IER_TFC 
                      |GMAC_IER_TCOMP
                      |GMAC_IER_ROVR 
                      |GMAC_IER_HRESP
                      |GMAC_IER_PFNZ 
                      |GMAC_IER_PTZ 
                      |GMAC_IER_PFTR
                      |GMAC_IER_EXINT
                      |GMAC_IER_DRQFR
                      |GMAC_IER_SFR
                      |GMAC_IER_DRQFT
                      |GMAC_IER_SFT
                      |GMAC_IER_PDRQFR
                      |GMAC_IER_PDRSFR
                      |GMAC_IER_PDRQFT
                      |GMAC_IER_PDRSFT;

    return;
}

// stop the MAC
STATIC VOID
MacStop( PENET_INFO pEthNetInfo )
{
//    UINT32  Temp;
//
//    // interrupt disable
//    //InterruptDisable( pEthNetInfo->nINTVirt );
//    ////pIntc->mask |= (1<<pEthNetInfo->nINT);
//
//    // stop Tx/Rx
//    Temp = MAC_REG(pEthNetInfo, MAC_MODE );
//    Temp &= ~(MAC_MODE_RX_ENABLE | MAC_MODE_TX_ENABLE);
//    MAC_REG( pEthNetInfo, MAC_MODE ) = Temp;

    return;
}



STATIC VOID
InitRxMem( PENET_INFO pEthNetInfo )
{
    UINT32          i;
    UINT32          Addr;    

DEBUGMSG(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pRxBuffer = 0x%x\n", pEthNetInfo->pRxBuffer);
DEBUGMSG(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pRxTd = 0x%x\n", pEthNetInfo->pRxTd);

    // initialize Rx releated memory 
    Addr = pEthNetInfo->pRxBuffer;
    for (i = 0; i < GRX_BUFFERS; i++)
    {
        GMACRXDESC   *pRxTd;
        pRxTd = pEthNetInfo->pRxTd + i;

        pRxTd->Addr.Val = Addr;  // physical address
        pRxTd->Status.Val = 0;
        Addr += GMAC_RX_UNITSIZE;
    }
    pEthNetInfo->pRxTd[(GRX_BUFFERS - 1)].Addr.Val |= GMAC_RX_WRAP_BIT;
    pEthNetInfo->RxTail = 0;

    return;
}

STATIC VOID
InitTxMem( PENET_INFO pEthNetInfo )
{
    UINT32          i;
    UINT32          Addr;    

DEBUGMSG(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pTxBuffer = 0x%x\n", pEthNetInfo->pTxBuffer);
DEBUGMSG(LAN_DBG_LVL,"LANInitMem : pEthNetInfo->pTxTd = 0x%x\n", pEthNetInfo->pTxTd);

    // initialize Tx releated memory 
    Addr = pEthNetInfo->pTxBuffer;
    for (i = 0; i < GTX_BUFFERS; i++)
    {
        GMACTXDESC   *pTxTd;
        pTxTd = pEthNetInfo->pTxTd + i;

        pTxTd->Addr = Addr;
        pTxTd->Status.Val = (UINT32)GMAC_TX_USED_BIT;
        Addr += GMAC_TX_UNITSIZE;
    }
    pEthNetInfo->pTxTd[(GTX_BUFFERS - 1)].Status.Val |= GMAC_TX_USED_BIT | GMAC_TX_WRAP_BIT;
    pEthNetInfo->TxHead = pEthNetInfo->TxTail = 0;

    return;
}



STATIC
VOID LANInitMem( PENET_INFO pEthNetInfo )
{
    UINT32          i;
    UINT32          Addr;    

    Addr = (UINT32)MemPageAlloc ((DMA_GMACB_SIZE + MEMPOOL_SIZE-1)>>MEMPOOL_SHIFT);

    Addr += DMA_ADDR_OFFSET;

    pEthNetInfo->pRxBuffer = Addr;       Addr += GRX_BUFFERS*GMAC_RX_UNITSIZE;
    pEthNetInfo->pTxBuffer = Addr;       Addr += GTX_BUFFERS*GMAC_TX_UNITSIZE;
    pEthNetInfo->pRxTd = Addr;           Addr += GRX_BUFFERS*MACB_DMADESC_SIZE;
    pEthNetInfo->pTxTd = Addr;           Addr += GTX_BUFFERS*MACB_DMADESC_SIZE;

    InitRxMem(pEthNetInfo);
    InitTxMem(pEthNetInfo);

    return;
}



// This is called to initialze the ethernet low level driver.  The base address of the ethernet hardware
// is passed into the routine.  The routine will return TRUE for a successful initialization.
STATIC INT32
LANInit( PENET_INFO pEthNetInfo, NET_PDEV pNetDev )
{
    //UINT32  MasterClock;
    //UINT32  Config;


	AT91PS_PMC	pPMC;
	//AT91PS_RSTC pRSTC;


	pPMC = (AT91PS_PMC)AT91C_BASE_PMC;
	//pRSTC= (AT91PS_RSTC)AT91C_BASE_RSTC;

    /* step 1 : enable peripheral */
	// Enable Peripheral clock in PMC for  EMAC		
	pPMC->PMC_PCER1 = ((unsigned int) 1 << (SAMA5D3_ID_GMAC-32));

	// The reset of the external chips is done in the beginning of the HALInit().	
	//pio_setup(hw_pio2, sizeof(hw_pio2)/sizeof(struct pio_desc));
	//pRSTC->RSTC_RCR = (0xA5 << 24 ) | AT91C_RSTC_EXTRST;

    /* step 2 : initialize the GMAC Device */
    pGMAC->GMAC_NCR = 0;                    /* disable the TX , RX and others */
    pGMAC->GMAC_IDR = ~0;                   /* Disable interrupts */
    pGMAC->GMAC_NCR |=  GMAC_NCR_CLRSTAT;   /* Clear all statistics registers */
    
    /* Clear receive status */
    pGMAC->GMAC_RSR = GMAC_RSR_RXOVR | GMAC_RSR_REC | GMAC_RSR_BNA |GMAC_RSR_HNO;
    /* Clear transmit status */
    pGMAC->GMAC_TSR = GMAC_TSR_UBR | GMAC_TSR_COL | GMAC_TSR_RLE
                      | GMAC_TSR_TXGO | GMAC_TSR_TFC | GMAC_TSR_TXCOMP
                      | GMAC_TSR_UND | GMAC_TSR_HRESP | GMAC_TSR_LCO;

    /* Clear interrupts */
	{ VOLATILE UINT32 dummy = pGMAC->GMAC_ISR;}


    /* Enable the copy of data into the buffers
       ignore broadcasts, and don't copy FCS. */
#define GMAC_CAF_DISABLE  0
#define GMAC_CAF_ENABLE   1
#define GMAC_NBC_DISABLE  0
#define GMAC_NBC_ENABLE   1
    if( GMAC_CAF_ENABLE )
    {
        pGMAC->GMAC_NCFGR = GMAC_NCFGR_FD | GMAC_NCFGR_GBE | 
                            GMAC_NCFGR_DBW_DBW64 | GMAC_NCFGR_CLK_MCK_64 |
                            GMAC_NCFGR_CAF ;
    }
    if( GMAC_NBC_DISABLE )
    {
        pGMAC->GMAC_NCFGR = GMAC_NCFGR_FD | GMAC_NCFGR_GBE |
                            GMAC_NCFGR_DBW_DBW64 | GMAC_NCFGR_CLK_MCK_64 |
                            GMAC_NCFGR_NBC ;
    }

    /* step 3 : initialize the GMAC Device */




    // interrupt for external 
    //AT91C_PIO_PE31


    //memset( EthNetInfo, 0, sizeof( ENET_INFO ) );

    pEthNetInfo->nINTPhys = SAMA5D3_ID_GMAC;
    //pEthNetInfo->nINTVirt = SYSINTR_ENET0;
    pEthNetInfo->PhyAddr = 0x1;
    //pEthNetInfo->PhyType = PHY_KS8737;
    //pEthNetInfo->PhyIntfMode = MAC_MODE_MII;
    //pEthNetInfo->RegBase = MAC0_BASE_ADDR;
    //
    ////pEthNetInfo->EthDev.nSCCDev = 0;
    //pEthNetInfo->EthDev.nTxBufDesc = TX_BD_DEFAULT;
    //pEthNetInfo->EthDev.nRxBufDesc = RX_BD_DEFAULT;
    //pEthNetInfo->EthDev.pTxBufDesc = NULL;
    //pEthNetInfo->EthDev.pRxBufDesc = NULL;
    //pEthNetInfo->EthDev.TxBufSize = FRAME_MAX_AL;
    //pEthNetInfo->EthDev.RxBufSize = FRAME_MAX_AL;

    /* atmel test program mac address is assigned */
    pNetDev->EthAddr[0] = 0x3a;
    pNetDev->EthAddr[1] = 0x1f;
    pNetDev->EthAddr[2] = 0x34;
    pNetDev->EthAddr[3] = 0x08;
    pNetDev->EthAddr[4] = 0x54;
    pNetDev->EthAddr[5] = 0x54; 


    pNetDev->BrcAddr[0] = 0xFF;
    pNetDev->BrcAddr[1] = 0xFF;
    pNetDev->BrcAddr[2] = 0xFF;
    pNetDev->BrcAddr[3] = 0xFF;
    pNetDev->BrcAddr[4] = 0xFF;
    pNetDev->BrcAddr[5] = 0xFF; 

    //DEBUGMSG(LAN_DBG_LVL,"%s: ", dev->name);

//    // enable clock
//    MasterClock = BOARD_MCK;
//    if( MasterClock <= 20000000 )
//        Config = MACB_BF(CLK, MACB_CLK_DIV8);
//    else
//    if (MasterClock <= 40000000)
//        Config = MACB_BF(CLK, MACB_CLK_DIV16);
//    else
//    if (MasterClock <= 80000000)
//        Config = MACB_BF(CLK, MACB_CLK_DIV32);
//    else
//        Config = MACB_BF(CLK, MACB_CLK_DIV64);
//
//    pGMAC->EMAC_NCFGR = Config;
//
//
//    if( pEthNetInfo->PhyIntfMode == MAC_MODE_MII )
//        pGMAC->EMAC_USRIO = MACB_BIT(CLKEN);
//    else
//        pGMAC->EMAC_USRIO = MACB_BIT(RMII) | MACB_BIT(CLKEN);




    // buffer description A ACN  A
    DEBUG1(LAN_DBG_LVL,"LANInitMem\n");
    LANInitMem( pEthNetInfo );

    // reset MAC
    DEBUG1(LAN_DBG_LVL,"MacReset\n");
    MacReset( pEthNetInfo );

    // init MAC
    DEBUG1(LAN_DBG_LVL,"MacInit\n");
    MacInit( pEthNetInfo, pNetDev );


{
    SYSINTR_REQUEST     SysIntrRequest;
    UINT32              nSysIntr = 0;

    DEBUGMSG( LAN_DBG_LVL,"%s (%d): ++\n", __FUNCTION__, __LINE__ );

    // request sysintr    
    SysIntrRequest.nHalIrq = SAMA5D3_ID_GMAC;
    SysIntrRequest.BitMap = 0;
    SysIntrRequest.Flag = SYSINTR_REQ_FLAG_STATIC;

    if( !KernelIoControl( IOCTRL_HAL_SYSINTR_REQUEST, &SysIntrRequest, sizeof( SYSINTR_REQUEST ), &nSysIntr, sizeof(UINT32) ) )
    {
        DEBUGMSG( LAN_DBG_LVL,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
        return 0;
    }
    DEBUGMSG( LAN_DBG_LVL,"%s (%d): nSysIntr = %d\n", __FUNCTION__, __LINE__, nSysIntr );
	
    pEthNetInfo->nINTVirt = nSysIntr;
}


//	Sleep(200);

    /* set mode */
	pio_setup(hw_pio_run, sizeof(hw_pio_run)/sizeof(struct pio_desc));
    Sleep(100);

    // setup PHY mode
    DEBUG1(LAN_DBG_LVL,"PhyInit\n");
    PhyInit( pEthNetInfo );

    //DEBUG1(LAN_DBG_LVL,"PhySetup\n");
    //PhySetup( pEthNetInfo );

    /* set run */

//	{ DWORD dummy = pGMAC->EMAC_ISR;}
//	pGMAC->EMAC_IER = 0x0000080;
//	{ DWORD dummy = pGMAC->EMAC_ISR;}


    return 0;

}   // LANInit



// This is called to initialze the ethernet low level driver.  The base address of the ethernet hardware
// is passed into the routine.  The routine will return TRUE for a successful initialization.
STATIC BOOL
LANOpen( PENET_INFO pEthNetInfo )
{
    UINT32      PhyMode;
    UINT32      MacMode;


    // LATER INITIAILIZE THE CARRIER DETECTION and SETUP THE PHY & MAC

//    // read PHY link status
//    PhyMode = PhyLinkModeRead( pEthNetInfo );
//
//    MacMode = 0x00020630 | (MAC_MODE_RX_ALL | MAC_MODE_MII | PhyMode
//#if (ENABLE_LOOPBACK & MAC_LOOPBACK)
//               | MAC_MODE_LOOPBACK | MAC_MODE_FULLDUPLEX
//#elif (ENABLE_LOOPBACK & PHY_LOOPBACK)
//               | MAC_MODE_FULLDUPLEX
//#endif
//                    );
//
//    if( pEthNetInfo->PhyIntfMode == MAC_MODE_RMII)
//    {
//        MacMode |= MAC_MODE_RMII;
//    }
//
//    // set MAC_MODE
//    MAC_REG( pEthNetInfo, MAC_MODE ) = MacMode;
//
//    // raise the interface flags - mark the device as up
//    PhyMode = PhyLinkStatusRead( pEthNetInfo );
//      if(PhyMode)
//    {
//        /* Link UP */
//        //netif_carrier_on (dev);
//    }
//      else
//    {
//        /* Link Down */
//        //netif_carrier_off (dev);
//    }

    MacStart( pEthNetInfo );

    return TRUE;
}


STATIC BOOL
LANClose( PENET_INFO pEthNetInfo )
{

    //netif_stop_queue (dev);

    // stop the device
    MacStop( pEthNetInfo );

    // disable interrupt
    InterruptDisable( pEthNetInfo->nINTVirt);

    return 0;
}





STATIC BOOL
LANSend( PENET_INFO pEthNetInfo )
{
    UINT32          Entry;
    UINT32          Ctrl;
    UINT32          Length;
    GMACTXDESC     *pTxTd;
    PBYTE           pPacketBuf;

    GmacStats *pGmacstatistics = pEthNetInfo->pGmacStats;;

DEBUGMSG(LAN_DBG_LVL,"%s(%d): \n",__FUNCTION__,__LINE__);

    /* If no free TxTd, buffer can't be sent */
    if( GCIRC_SPACE(pEthNetInfo->TxHead, pEthNetInfo->TxTail, GTX_BUFFERS) == 0)
    {
        //netif_stop_queue(dev);
        return FALSE;
    }

    Entry = pEthNetInfo->TxHead;
DEBUGMSG(LAN_DBG_LVL,"%s(%d): + : Entry = %d \n",__FUNCTION__,__LINE__, Entry);

    /* Pointers to the current TxTd */
    pTxTd = &pEthNetInfo->pTxTd[Entry];

//    pPacketBuf = pEthNetInfo->pTxBuffer + Entry*GMAC_TX_UNITSIZE;
    pPacketBuf = (PBYTE)pTxTd->Addr;

    // get packet data from network server
    if( !NwPacketOutput( pEthNetInfo->iNetIf, pPacketBuf, &Length ) )
        return FALSE;

    // check the length
    Length = max( FRAME_MIN - 4, Length );
    Length = min( FRAME_MAX - 4, Length );


//// print the transmitted packet for debug
//{
//    int i; 
//    for(i = 0; i<Length; i++) 
//        DEBUGMSG(LAN_DBG_LVL,"%2.2x ",*(pPacketBuf + i) );
//    DEBUGMSG(LAN_DBG_LVL,"\n");
//}


DEBUGMSG(LAN_DBG_LVL,"%s(%d): pTxTd = %x\n",__FUNCTION__,__LINE__, pTxTd);
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pTxTd->Addr = %x\n",__FUNCTION__,__LINE__, pTxTd->Addr);
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pTxTd->Status.Val = %x\n",__FUNCTION__,__LINE__, pTxTd->Status.Val);

    /* Update TD status. The buffer size defined is length of ethernet frame
      so it's always the last buffer of the frame. */
    if (pEthNetInfo->TxHead == GTX_BUFFERS-1)
    {
        pTxTd->Status.Val = (Length & GMAC_LENGTH_FRAME) | GMAC_TX_LAST_BUFFER_BIT | 
                            GMAC_TX_WRAP_BIT;
    }
    else
    {
        pTxTd->Status.Val = (Length & GMAC_LENGTH_FRAME) | GMAC_TX_LAST_BUFFER_BIT;
    }

    GCIRC_INC( pEthNetInfo->TxHead, GTX_BUFFERS );
DEBUGMSG(LAN_DBG_LVL,"%s(%d): : pTxTd = %x \n",__FUNCTION__,__LINE__, pTxTd);
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pTxTd->Addr = %x\n",__FUNCTION__,__LINE__, pTxTd->Addr);
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pTxTd->Status.Val = %x\n",__FUNCTION__,__LINE__, pTxTd->Status.Val);

    /* Tx packets count */
    pGmacstatistics->tx_packets++;

    pGMAC->GMAC_NCR |= GMAC_NCR_TSTART;
DEBUGMSG(LAN_DBG_LVL,"%s(%d): : pGMAC->GMAC_NCR = %x \n",__FUNCTION__,__LINE__, pGMAC->GMAC_NCR);
//    if( TX_BUFFS_AVAIL(pEthNetInfo) < 1 )
//        netif_stop_queue(dev);

    return TRUE;
}


STATIC VOID
LANSendDone( PENET_INFO pEthNetInfo )
{
    GMACTXDESC      *pTxTd;

    if (!GCIRC_EMPTY(pEthNetInfo->TxHead, pEthNetInfo->TxTail))
    {
        /* Check the buffers */
        do
        {
            pTxTd = &pEthNetInfo->pTxTd[pEthNetInfo->TxTail];
            //pTxCb = &pGmacd->fTxCbList[pGmacd->wTxTail];
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pEthNetInfo->TxTail = %d\n",__FUNCTION__,__LINE__, pEthNetInfo->TxTail);

            /* Exit if buffer has not been sent yet */
            if ((pTxTd->Status.Val & (UINT32)GMAC_TX_USED_BIT) == 0)
            {
                break;
            }

            ///* Notify upper layer that a packet has been sent */
            //if (*pTxCb)
            //{
            //    (*pTxCb)(txStatusFlag);
            //}

            GCIRC_INC( pEthNetInfo->TxTail, GTX_BUFFERS );
        } while (GCIRC_CNT(pEthNetInfo->TxHead, pEthNetInfo->TxTail, GTX_BUFFERS));
    }
DEBUGMSG(LAN_DBG_LVL,"%s(%d): pEthNetInfo->TxTail = %d\n",__FUNCTION__,__LINE__, pEthNetInfo->TxTail);

    return;
}
        //    UINT32  Status;
//    UINT32  Head, Tail;
//
//    Status = pGMAC->EMAC_TSR;
//
//    pGMAC->EMAC_TSR = ~0;
//
//DEBUGMSG(LAN_DBG_LVL,"LANSendDone :Status = %x \n", Status);
//
//    if( Status & MACB_BIT(UND) )
//    {
//        // reset the buffer
//        INT32           i;
//        PMACB_DMADESC   pRing;
//
//        for (i = 0; i < TX_RING_SIZE; i++)
//        {
//            pRing = pEthNetInfo->pTxRing + i;
//    
//            pRing->Addr = 0;
//            pRing->Ctrl = MACB_BIT(TX_USED);
//        }
//        pRing = pEthNetInfo->pTxRing + (TX_RING_SIZE - 1);
//        pRing->Ctrl |= MACB_BIT(TX_WRAP);
//    
//        pEthNetInfo->TxHead = pEthNetInfo->TxTail = 0;
//
//    }
//
//    if( !(Status & MACB_BIT(COMP)) )
//        return;
//
//    Head = pEthNetInfo->TxHead;
//    for( Tail = pEthNetInfo->TxTail; Tail != Head; Tail = NEXT_TX(Tail) )
//    {
//        //PENET_RINGINFO  pRingInfo = pEthNetInfo->TxRingInfo;
//        PMACB_DMADESC   pTxRing;
//        UINT32          BufState;
//
//        pTxRing = pEthNetInfo->pTxRing + Tail;
//        BufState = pTxRing->Ctrl;
//        if( !(BufState &  MACB_BIT(TX_USED)) )
//            break;
//
//        pEthNetInfo->Stat.nTxPacket++;
//        //pEthNetInfo->Stat.nTxByte += Len;
//        //bp->stats.tx_packets++;
//        //bp->stats.tx_bytes += skb->len;
//        //rp->skb = NULL;
//        //dev_kfree_skb_irq(skb);
//    }
//    
//    pEthNetInfo->TxTail = Tail;
//    //if( netif_queue_stopped(bp->dev) && (TX_BUFFS_AVAIL(bp) > MACB_TX_WAKEUP_THRESH) )
//    //    netif_wake_queue(bp->dev);
//DEBUGMSG(LAN_DBG_LVL,"LANSendDone : Tail = %d \n", Tail);
//
//    return;
//}


STATIC INT32
MACB_RxFrame( PENET_INFO pEthNetInfo, UINT32 FirstFrag, UINT32 LastFrag )
{
    UINT32          Length;
    UINT32          Frag;
    UINT32          Offset;
    PBYTE           pRxBuffer;
    GMACRXDESC     *pRxTd;

//DEBUGMSG(LAN_DBG_LVL,"MACB_RxFrame : FirstFrag= %d\n", FirstFrag);
//DEBUGMSG(LAN_DBG_LVL,"MACB_RxFrame : LastFrag = %d\n", LastFrag);

    pRxTd = pEthNetInfo->pRxTd + LastFrag;
    Length = (pRxTd->Status.Val & GMAC_LENGTH_FRAME);

    pRxBuffer = pEthNetInfo->RxPacket + RX_OFFSET;
    Offset = 0;
DEBUGMSG(LAN_DBG_LVL,"MACB_RxFrame : pRxBuffer = %x\n", pRxBuffer);
DEBUGMSG(LAN_DBG_LVL,"MACB_RxFrame : Length    = %d\n", Length);
    for( Frag = FirstFrag; ; GCIRC_INC(Frag, GRX_BUFFERS) )
    {
        UINT32  FragLen = GMAC_RX_UNITSIZE;

        if( Offset + FragLen > Length )
        {
            FragLen = Length - Offset;
        }

        MemCpy( pRxBuffer + Offset, pEthNetInfo->pRxBuffer + (Frag<<GMAC_RX_UNITSIZE_SHIFT), FragLen );
        Offset += GMAC_RX_UNITSIZE;

        pRxTd = pEthNetInfo->pRxTd + Frag;
        pRxTd->Addr.Val &= ~(GMAC_RX_OWNERSHIP_BIT);

        if( Frag == LastFrag )
            break;
    }
//// print the received packet for debug
//{
//    int i; 
//    for(i = 0; i<Length; i++) 
//        DEBUGMSG(LAN_DBG_LVL,"%2.2x ",*(pRxBuffer + i) );
//    DEBUGMSG(LAN_DBG_LVL,"\n");
//}
    
    NwPacketInput( pEthNetInfo->iNetIf, pRxBuffer, Length );

    //RxPackets++;
    //RxBytes += Length;


//    pEthNetInfo->Stat.nRxPacket++;
//    pEthNetInfo->Stat.nRxByte += Length;

//DEBUGMSG(LAN_DBG_LVL,"MACB_RxFrame : pEthNetInfo->Stat.nRxPacket = %d\n", pEthNetInfo->Stat.nRxPacket);
    //netif_receive_skb(skb);

    return 0;
}


/* Mark DMA descriptors from begin up to and not including end as unused */
STATIC VOID
DiscardPartialFrame(PENET_INFO pEthNetInfo, UINT32 FirstFrag, UINT32 LastFrag)
{
    GMACRXDESC   *pRxTd;

//DEBUGMSG(LAN_DBG_LVL,"DiscardPartialFrame: %d - %d\n",FirstFrag, LastFrag );
    while (FirstFrag != LastFrag)
    {
        pRxTd = pEthNetInfo->pRxTd + FirstFrag;
        pRxTd->Addr.Val &= ~(GMAC_RX_OWNERSHIP_BIT);
        GCIRC_INC(FirstFrag, GRX_BUFFERS);
    }


    /*
     * When this happens, the hardware stats registers for
     * whatever caused this is updated, so we don't have to record
     * anything.
     */

    return;
}

STATIC UINT32 MACB_Rx( PENET_INFO pEthNetInfo )
{
    UINT32  Addr, Ctrl;
    UINT32  ReceivedPacket = 0;
    INT32   iSOF = -1;

    UINT32  Tail = pEthNetInfo->RxTail;
    GMACRXDESC  *pRxTd;

    GmacStats *pGmacstatistics = pEthNetInfo->pGmacStats;;

    for( ; ; Tail = GCIRC_INC(Tail, GRX_BUFFERS) )
    {
        pRxTd = pEthNetInfo->pRxTd + Tail;

        if(!((pRxTd->Addr.Val & GMAC_RX_OWNERSHIP_BIT) == GMAC_RX_OWNERSHIP_BIT))
            break;

        /* A start of frame has been received, discard previous fragments */        
        if ((pRxTd->Status.Val & GMAC_RX_SOF_BIT) == GMAC_RX_SOF_BIT)       
        {
            if( Tail != pEthNetInfo->RxTail )
            {
                /* discard previous fragment */
                DiscardPartialFrame(pEthNetInfo, pEthNetInfo->RxTail, Tail);

            }
            /* Start to gather buffers in a frame */
            iSOF = Tail;
        }

        /* An end of frame has been received, return the data */
        if ((pRxTd->Status.Val & GMAC_RX_EOF_BIT) == GMAC_RX_EOF_BIT)
        {
            INT32   Dropped;
//DEBUGMSG(LAN_DBG_LVL,"MACB_Rx : iSOF, Tail = %d - %d\n", iSOF, Tail);
            Dropped = MACB_RxFrame(pEthNetInfo, iSOF, Tail);
            iSOF = -1;
            if( !Dropped )
            {
                ReceivedPacket++;
            }
        }
    }

    if( iSOF != -1 )
        pEthNetInfo->RxTail = iSOF;
    else
        pEthNetInfo->RxTail = Tail;



////        /* Increment the pointer */
////        GCIRC_INC(Tail, GRX_BUFFERS);
////        asm("nop");
//
//        /* Copy data in the frame buffer */
//        if (isFrame)
//        {
//            if (Tail == pEthNetInfo->RxTail)
//            {
//                kprintf("no EOF (Invalid of buffers too small)\n\r");
//
//                do
//                {
//                    pRxTd = pEthNetInfo->pRxTd + pEthNetInfo->RxTail;
//                    pRxTd->Addr.Val &= ~(GMAC_RX_OWNERSHIP_BIT);
//                    GCIRC_INC(pEthNetInfo->RxTail, GRX_BUFFERS);
//                } while(Tail != pEthNetInfo->RxTail);
//                return GMACD_RX_NULL;
//            }
//
//            /* Copy the buffer into the application frame */
//            bufferLength = GMAC_RX_UNITSIZE;
//            if ((tmpFrameSize + bufferLength) > frameSize)
//            {
//                bufferLength = frameSize - tmpFrameSize;
//            }
//            memcpy(pTmpFrame, (void*)(pRxTd->addr.val & GMAC_ADDRESS_MASK), bufferLength);
//            pTmpFrame += bufferLength;
//            tmpFrameSize += bufferLength;
//
//            /* An end of frame has been received, return the data */
//            if ((pRxTd->Status.Val & GMAC_RX_EOF_BIT) == GMAC_RX_EOF_BIT)
//            {
//                /* Frame size from the GMAC */
//                *pRcvSize = (pRxTd->Status.Val & GMAC_LENGTH_FRAME);
//
//                /* Application frame buffer is too small all data have not been copied */
//                if (tmpFrameSize < *pRcvSize) {
//                    return GMACD_SIZE_TOO_SMALL;
//                }
//                TRACE_DEBUG("packet %d-%d (%d)\n\r", pGmacd->wRxI, tmpIdx, *pRcvSize);
//                /* All data have been copied in the application frame buffer => release TD */
//                while (pGmacd->wRxI != tmpIdx)
//                {
//                    pRxTd = &pGmacd->pRxD[pGmacd->wRxI];
//                    pRxTd->addr.val &= ~(GMAC_RX_OWNERSHIP_BIT);
//                    GCIRC_INC(pGmacd->wRxI, pGmacd->wRxListSize);
//                }
//                pGmacstatistics->rx_comp++;
//                return GMACD_OK;
//            }
//        }
//        
//        /* SOF has not been detected, skip the fragment */
//        else
//        {
//           pRxTd->Addr.Val &= ~(GMAC_RX_OWNERSHIP_BIT);
//           pGmacd->wRxI = tmpIdx;
//        }
//
//
//        Addr = pRxRing->Addr;
//        Ctrl = pRxRing->Ctrl;
////DEBUGMSG(LAN_DBG_LVL,"MACB_Rx : Addr = %x\n", Addr);
////DEBUGMSG(LAN_DBG_LVL,"MACB_Rx : Ctrl = %x\n", Ctrl);
//        if( !(Addr & MACB_BIT(RX_USED)) )
//            break;
//
//        if( Ctrl & MACB_BIT(RX_SOF) )
//        {
//            if( iSOF != -1 )
//                DiscardPartialFrame(pEthNetInfo, iSOF, Tail);
//            iSOF = Tail;
//        }
//
//        if( Ctrl & MACB_BIT(RX_EOF) )
//        {
//            INT32   Dropped;
//
////DEBUGMSG(LAN_DBG_LVL,"MACB_Rx : iSOF, Tail = %d - %d\n", iSOF, Tail);
//            Dropped = MACB_RxFrame(pEthNetInfo, iSOF, Tail);
//            iSOF = -1;
//            if( !Dropped )
//            {
//                ReceivedPacket++;
//            }
//        }
//    }
//
//    if( iSOF != -1 )
//        pEthNetInfo->RxTail = iSOF;
//    else
//        pEthNetInfo->RxTail = Tail;


    return ReceivedPacket;
            
}



STATIC VOID
LANReceive( PENET_INFO pEthNetInfo, UINT32 Status )
{
//    UINT32  Status;
//
//    Status = pGMAC->EMAC_RSR;
//    pGMAC->EMAC_RSR = Status;
//DEBUGMSG(LAN_DBG_LVL,"LANReceive : EMAC_RSR = %x\n", Status);
//    if( !Status )
//    {
//        /*
//         * This may happen if an interrupt was pending before
//         * this function was called last time, and no packets
//         * have been received since.
//         */
//        //netif_rx_complete(dev);
//        goto OUT_LAN_RECEIVE;
//    }
//
//    if( !(Status & MACB_BIT(REC)) )
//    {
//        DEBUG1(LAN_DBG_LVL,"No RX buffers complete\n");
//        //netif_rx_complete(dev);
//        goto OUT_LAN_RECEIVE;
//    }

    MACB_Rx( pEthNetInfo );
    //work_done = MACB_Rx( pEthNetInfo );
    //if (work_done < orig_budget)
    //{
    //    netif_rx_complete(dev);
    //    retval = 0;
    //}
    //else
    //{
    //    retval = 1;
    //}

//    /*
//     * We've done what we can to clean the buffers. Make sure we
//     * get notified when new packets arrive.
//     */
//OUT_LAN_RECEIVE:
//    pGMAC->EMAC_IER = MACB_RX_INT_FLAGS;
    
    return;
}


STATIC VOID
LANISR( PENET_INFO pEthNetInfo )
{
    GmacStats *pGmacstatistics = pEthNetInfo->pGmacStats;;

    volatile UINT32 isr;
    volatile UINT32 rsr;
    volatile UINT32 tsr;
    UINT32 rxStatusFlag;
    UINT32 txStatusFlag;

    isr = pGMAC->GMAC_ISR;
    rsr = pGMAC->GMAC_RSR;
    tsr = pGMAC->GMAC_TSR;

    isr &= ~(pGMAC->GMAC_IMR | 0xF8030300);

    /* RX packet */
    if ((isr & GMAC_ISR_RCOMP) || (rsr & GMAC_RSR_REC))
    {
        rxStatusFlag = GMAC_RSR_REC;

        /* Frame received */
        pGmacstatistics->rx_packets++;
        /* Check OVR */
        if (rsr & GMAC_RSR_RXOVR)
        {
            rxStatusFlag |= GMAC_RSR_RXOVR;
            pGmacstatistics->rx_ovrs++;
        }
        /* Check BNA */
        if (rsr & GMAC_RSR_BNA)
        {
            rxStatusFlag |= GMAC_RSR_BNA;
             pGmacstatistics->rx_bnas++;
        }
        /* Check HNO */
        if (rsr & GMAC_RSR_HNO)
        {
            rxStatusFlag |= GMAC_RSR_HNO;
             pGmacstatistics->rx_ovrs++;
        }        
        /* Clear status */
        pGMAC->GMAC_RSR = rxStatusFlag;

        /* Invoke callbacks */
        //if (pGmacd->fRxCb)
        //{
            LANReceive(pEthNetInfo, rxStatusFlag);
            //pGmacd->fRxCb(rxStatusFlag);
        //}
    }

    /* TX packet */
    if ((isr & GMAC_ISR_TCOMP) || (tsr & GMAC_TSR_TXCOMP))
    {
DEBUGMSG(LAN_DBG_LVL,"%s(%d): TX COMP\n",__FUNCTION__,__LINE__);

DEBUGMSG(LAN_DBG_LVL,"%s(%d): isr = %x\n",__FUNCTION__,__LINE__, isr);
DEBUGMSG(LAN_DBG_LVL,"%s(%d): tsr = %x\n",__FUNCTION__,__LINE__, tsr);

        txStatusFlag = GMAC_TSR_TXCOMP;
        pGmacstatistics->tx_comp ++;

        /*  A frame transmitted Check RLE */
        if (tsr & GMAC_TSR_RLE)
        {
            /* Status RLE & Number of discarded buffers */
            txStatusFlag = GMAC_TSR_RLE
                         | GCIRC_CNT(pEthNetInfo->TxHead,
                                     pEthNetInfo->TxTail,
                                     GTX_BUFFERS);
            //pTxCb = &pGmacd->fTxCbList[pEthNetInfo->TxTail];
            //GMACD_ResetTx(pEthNetInfo);
            InitTxMem(pEthNetInfo);
            MacResetTx(pEthNetInfo);
            kprintf("Tx RLE!!\n\r");
            /* enable the Tx */
            pGMAC->GMAC_NCR |=  GMAC_NCR_TXEN;
            pGmacstatistics->tx_errors++;
        }
        /* Check COL */
        if (tsr & GMAC_TSR_COL)
        {
            txStatusFlag |= GMAC_TSR_COL;
            pGmacstatistics->collisions++;
        }
        /* Check TFC */
        if (tsr & GMAC_TSR_TFC)
        {
            txStatusFlag |= GMAC_TSR_TFC;
            pGmacstatistics->tx_exausts++;
        }
        /* Check UND */
        if (tsr & GMAC_TSR_UND)
        {
            txStatusFlag |= GMAC_TSR_UND;
            pGmacstatistics->tx_underruns++;
        }
        /* Check HRESP */
        if (tsr & GMAC_TSR_HRESP)
        {
            txStatusFlag |= GMAC_TSR_HRESP;
            pGmacstatistics->tx_hresp++;
        }
        /* Check LCO */
        if (tsr & GMAC_TSR_LCO)
        {
            txStatusFlag |= GMAC_TSR_LCO;
            pGmacstatistics->tx_latecollisions++;
        }        
        /* Clear status */
        pGMAC->GMAC_TSR = txStatusFlag;

        LANSendDone(pEthNetInfo);
        
        if (tsr & GMAC_TSR_RLE)
        {
            /* Notify upper layer RLE */
            //if (*pTxCb) {
            //    (*pTxCb)(txStatusFlag);
            //}
        }
//  AT91_REG      GMAC_OTLO;       /**< \brief (Gmac Offset: 0x100) Octets Transmitted [31:0] Register */
//  AT91_REG      GMAC_OTHI;       /**< \brief (Gmac Offset: 0x104) Octets Transmitted [47:32] Register */
//  AT91_REG      GMAC_FT;         /**< \brief (Gmac Offset: 0x108) Frames Transmitted Register */
//  AT91_REG      GMAC_BCFT;       /**< \brief (Gmac Offset: 0x10C) Broadcast Frames Transmitted Register */
//  AT91_REG      GMAC_MFT;        /**< \brief (Gmac Offset: 0x110) Multicast Frames Transmitted Register */
//  AT91_REG      GMAC_PFT;        /**< \brief (Gmac Offset: 0x114) Pause Frames Transmitted Register */
//  AT91_REG      GMAC_BFT64;      /**< \brief (Gmac Offset: 0x118) 64 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TBFT127;    /**< \brief (Gmac Offset: 0x11C) 65 to 127 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TBFT255;    /**< \brief (Gmac Offset: 0x120) 128 to 255 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TBFT511;    /**< \brief (Gmac Offset: 0x124) 256 to 511 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TBFT1023;   /**< \brief (Gmac Offset: 0x128) 512 to 1023 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TBFT1518;   /**< \brief (Gmac Offset: 0x12C) 1024 to 1518 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_GTBFT1518;  /**< \brief (Gmac Offset: 0x130) Greater Than 1518 Byte Frames Transmitted Register */
//  AT91_REG      GMAC_TUR;        /**< \brief (Gmac Offset: 0x134) Transmit Under Runs Register */
//  AT91_REG      GMAC_SCF;        /**< \brief (Gmac Offset: 0x138) Single Collision Frames Register */
//  AT91_REG      GMAC_MCF;        /**< \brief (Gmac Offset: 0x13C) Multiple Collision Frames Register */
//  AT91_REG      GMAC_EC;         /**< \brief (Gmac Offset: 0x140) Excessive Collisions Register */
//  AT91_REG      GMAC_LC;         /**< \brief (Gmac Offset: 0x144) Late Collisions Register */
//  AT91_REG      GMAC_DTF;        /**< \brief (Gmac Offset: 0x148) Deferred Transmission Frames Register */
//  AT91_REG      GMAC_CSE;        /**< \brief (Gmac Offset: 0x14C) Carrier Sense Errors Register */
//  AT91_REG      GMAC_ORLO;       /**< \brief (Gmac Offset: 0x150) Octets Received [31:0] Received */
//  AT91_REG      GMAC_ORHI;       /**< \brief (Gmac Offset: 0x154) Octets Received [47:32] Received */
//  AT91_REG      GMAC_FR;         /**< \brief (Gmac Offset: 0x158) Frames Received Register */
//  AT91_REG      GMAC_BCFR;       /**< \brief (Gmac Offset: 0x15C) Broadcast Frames Received Register */
//  AT91_REG      GMAC_MFR;        /**< \brief (Gmac Offset: 0x160) Multicast Frames Received Register */
//  AT91_REG      GMAC_PFR;        /**< \brief (Gmac Offset: 0x164) Pause Frames Received Register */
//  AT91_REG      GMAC_BFR64;      /**< \brief (Gmac Offset: 0x168) 64 Byte Frames Received Register */
//  AT91_REG      GMAC_TBFR127;    /**< \brief (Gmac Offset: 0x16C) 65 to 127 Byte Frames Received Register */
//  AT91_REG      GMAC_TBFR255;    /**< \brief (Gmac Offset: 0x170) 128 to 255 Byte Frames Received Register */
//  AT91_REG      GMAC_TBFR511;    /**< \brief (Gmac Offset: 0x174) 256 to 511Byte Frames Received Register */
//  AT91_REG      GMAC_TBFR1023;   /**< \brief (Gmac Offset: 0x178) 512 to 1023 Byte Frames Received Register */
//  AT91_REG      GMAC_TBFR1518;   /**< \brief (Gmac Offset: 0x17C) 1024 to 1518 Byte Frames Received Register */
//  AT91_REG      GMAC_TMXBFR;     /**< \brief (Gmac Offset: 0x180) 1519 to Maximum Byte Frames Received Register */
//  AT91_REG      GMAC_UFR;        /**< \brief (Gmac Offset: 0x184) Undersize Frames Received Register */
//  AT91_REG      GMAC_OFR;        /**< \brief (Gmac Offset: 0x188) Oversize Frames Received Register */
//  AT91_REG      GMAC_JR;         /**< \brief (Gmac Offset: 0x18C) Jabbers Received Register */
//  AT91_REG      GMAC_FCSE;       /**< \brief (Gmac Offset: 0x190) Frame Check Sequence Errors Register */
//  AT91_REG      GMAC_LFFE;       /**< \brief (Gmac Offset: 0x194) Length Field Frame Errors Register */
//  AT91_REG      GMAC_RSE;        /**< \brief (Gmac Offset: 0x198) Receive Symbol Errors Register */
//  AT91_REG      GMAC_AE;         /**< \brief (Gmac Offset: 0x19C) Alignment Errors Register */
//  AT91_REG      GMAC_RRE;        /**< \brief (Gmac Offset: 0x1A0) Receive Resource Errors Register */
//  AT91_REG      GMAC_ROE;        /**< \brief (Gmac Offset: 0x1A4) Receive Overrun Register */
//  AT91_REG      GMAC_IHCE;       /**< \brief (Gmac Offset: 0x1A8) IP Header Checksum Errors Register */
//  AT91_REG      GMAC_TCE;        /**< \brief (Gmac Offset: 0x1AC) TCP Checksum Errors Register */
//  AT91_REG      GMAC_UCE;        /**< \brief (Gmac Offset: 0x1B0) UDP Checksum Errors Register */



//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_OTLO = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_OTLO);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_OTHI = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_OTHI);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_FT = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_FT);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_PFT = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_PFT);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_BFT64 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_BFT64);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBFT127 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBFT127);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBFT255 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBFT255);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBFT511 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBFT511);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBFT1023 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBFT1023);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBFT1518 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBFT1518);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_GTBFT1518 = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_GTBFT1518);
//
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_AE = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_AE);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_RRE = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_RRE);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_IHCE = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_IHCE);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TCE = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TCE);
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_UCE = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_UCE);
//
//DEBUGMSG(LAN_DBG_LVL,"%s(%d): pGMAC->GMAC_TBQB = %x\n",__FUNCTION__,__LINE__, pGMAC->GMAC_TBQB);
        
        
        ///* If a wakeup has been scheduled, notify upper layer that it can
        //   send other packets, send will be successfull. */
        //if((GCIRC_SPACE( pEthNetInfo->TxHead,
        //                 pEthNetInfo->TxTail,
        //                 GTX_BUFFERS) >= pGmacd->bWakeupThreshold) && pGmacd->fWakupCb)
        //{
        //    pGmacd->fWakupCb();
        //}
    }

    /* PAUSE Frame */
    if (isr & GMAC_ISR_PFNZ) kprintf("Pause!\n\r");
    if (isr & GMAC_ISR_PTZ)  kprintf("Pause TO!\n\r");





//
//    VOLATILE UINT32  IntStatus;
//    
//    IntStatus = pGMAC->EMAC_ISR;
//    if( IntStatus == 0 )
//        return;
//
//    while( IntStatus )
//    {
//        //if( IntStatus & MACB_BIT(MFD) )
//        //    Complete(&pEthNetInfo->mdio_complete);
//
//        //if(NetIfRunning())
//        //    pGMAC->EMAC_IDR = ~0;
//
//        // Rx
//        if( IntStatus & MACB_RX_INT_FLAGS )
//        {
//            //DEBUGMSG(LAN_DBG_LVL,"LANISR : LANReceive = %x\n", nRxDone );
//            LANReceive( pEthNetInfo );
//        }
//        
//        // Tx done
//        if( IntStatus & (MACB_BIT(TCOMP) | MACB_BIT(ISR_TUND)) )
//        {
//            //DEBUGMSG(LAN_DBG_LVL,"LANISR : LANSendDone = %x\n", nTxDone );
//            LANSendDone( pEthNetInfo );
//        }
//
//        // remained op
//        if( IntStatus & MACB_BIT(HRESP) )
//        {
//            /*
//            * TODO: Reset the hardware, and maybe move the printk
//            * to a lower-priority context as well (work queue?)
//            */
//            DEBUG1(LAN_DBG_LVL,"DMA bus error: HRESP not OK\n");
//        }
//
//        // re-read the status register for next interrupt
//        IntStatus = pGMAC->EMAC_ISR;
//    }

    return;
}


//    struct net_device *dev = dev_id;
//    struct macb *bp = netdev_priv(dev);
//    u32 status;
//
//    status = macb_readl(bp, ISR);
//
//    if (unlikely(!status))
//        return IRQ_NONE;
//
//    spin_lock(&bp->lock);
//
//    while (status) {
//        if (status & MACB_BIT(MFD))
//            complete(&bp->mdio_complete);
//
//        /* close possible race with dev_close */
//        if (unlikely(!netif_running(dev))) {
//            macb_writel(bp, IDR, ~0UL);
//            break;
//        }
//
//        if (status & MACB_RX_INT_FLAGS) {
//            if (netif_rx_schedule_prep(dev)) {
//                /*
//                 * There's no point taking any more interrupts
//                 * until we have processed the buffers
//                 */
//                macb_writel(bp, IDR, MACB_RX_INT_FLAGS);
//                dev_dbg(&bp->pdev->dev, "scheduling RX softirq\n");
//                __netif_rx_schedule(dev);
//            }
//        }
//
//        if (status & (MACB_BIT(TCOMP) | MACB_BIT(ISR_TUND)))
//            macb_tx(bp);
//
//        /*
//         * Link change detection isn't possible with RMII, so we'll
//         * add that if/when we get our hands on a full-blown MII PHY.
//         */
//
//        if (status & MACB_BIT(HRESP)) {
//            /*
//             * TODO: Reset the hardware, and maybe move the printk
//             * to a lower-priority context as well (work queue?)
//             */
//            printk(KERN_ERR "%s: DMA bus error: HRESP not OK\n",
//                   dev->name);
//        }
//
//        status = macb_readl(bp, ISR);
//    }
//
//    spin_unlock(&bp->lock);
//
//    return IRQ_HANDLED;
//    
//    
//    
//    
//
//
//    UINT32          i;
//    UINT            IntStatus, IntEnabled;
//    PENET_BUFDESC   pBufferToProcess = NULL;
//    PENET_BUFDESC   pBufferProcessed = NULL;
//
//    UINT32          nRxDone, nRxNotTaken, TxFifoLevel;
//    VOLATILE UINT32 Temp;
//
//    // read device interrupt status
//    MAC_REG_READ( pEthNetInfo, INT_SRC, IntStatus);
//    //MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus);
//    MAC_REG_READ( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, IntEnabled );
//    IntStatus &= IntEnabled;
//
////DEBUGMSG(LAN_DBG_LVL,"LANISR : IntEnabled = %x\n", IntEnabled );
////DEBUGMSG(LAN_DBG_LVL,"LANISR : IntStatus = %x\n", IntStatus );
//    // quick exit if device did not generate interrupt
//    if( IntStatus == 0)
//    {
//        goto INT_RETURN;
//        //InterruptDone( SYSINTR_LAN );
//        //return;
//    }
//
//    // handle receiver interrupt
//    if( IntStatus & (MAC_INT_RX_DONE|MAC_INT_RX_FRAME_CUT) )
//    {
////DEBUGMSG(LAN_DBG_LVL,"LANISR : MAC_INT_RX_DONE|MAC_INT_RX_FRAME_CUT\n" );
//        if( IntStatus & MAC_INT_RX_DONE )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_DONE);
//        }
//        
//        if( IntStatus & MAC_INT_RX_FRAME_CUT )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_CUT );
//        }
//
//        // check Rx buffer level
//        MAC_REG_READ( pEthNetInfo, RX_BUFLVL, Temp );
//        nRxDone = Temp & 0x3F;
//        nRxNotTaken = (Temp & 0x3F00) >> 8;
//
////DEBUGMSG(LAN_DBG_LVL,"LANISR : nRxDone = %x\n", nRxDone );
//        if (nRxDone == 0)
//        {
//            goto INT_RETURN;
//            //SYS_INT_ACK (priv->ivec);
//            //spin_unlock (&priv->lock);
//            //return;
//        }
//
//        for( i = 0; i < nRxDone; i++ )
//        {
//            // obtain the current buffer address
//            pBufferProcessed = pEthNetInfo->pNextRxDone;
//            
//            MAC_REG_READ(pEthNetInfo, RX_DONE_BADDR, Temp );
//            MAC_REG_READ(pEthNetInfo, RX_BSTAT, Temp );
//            pBufferProcessed->FrameLength = ((Temp >> 16) & 0xFFFF);
//            pBufferProcessed->FrameError = (Temp & 0x5);
//
//            pBufferProcessed->Status = _B_FILLED_;
//            pEthNetInfo->pNextRxDone = pBufferProcessed->pNext;
//        }
//
//        for (i = 0; i < MAC_RX_BADDR_REG_FIFO_DEPTH - nRxNotTaken; i++)
//        {
//            pBufferToProcess = pEthNetInfo->pNextDeviceRx;
//            if( pBufferToProcess->Status == _B_FREE_ )    // there exists free buffer(s)
//            {
//                MAC_REG_WRITE( pEthNetInfo, RX_BADDR, (UINT32) pBufferToProcess->Buf);
//                pEthNetInfo->pNextDeviceRx = pBufferToProcess->pNext;
//
//                //new_skb = dev_alloc_skb (FRAME_MAX_AL);
//                //
//                //if (new_skb != NULL)
//                //{
//                //    new_skb->data += PAYLOAD_ALIGN_OFFSET;
//                //    new_skb->tail += PAYLOAD_ALIGN_OFFSET;
//                //
//                //    bufferToProcess->skb = new_skb;
//                //    bufferToProcess->pCluster = new_skb->data - PAYLOAD_ALIGN_OFFSET;
//                //
//                //    /* link new buffer to the device -> it will awake the device */
//                //    MAC_REG_WRITE (priv, RX_BADDR, (UINT) bufferToProcess->pCluster);
//                //
//                //    /* update next processing pointer (advance the pointer) */
//                //    priv->nextDeviceRx = bufferToProcess->next;
//                //}
//                //else
//                //{
//                //    break;
//                //}
//            }
//            else            /* no free buffer -> reuse the current buffer */
//            {
//                break;
//            }
//        }
//
//        /* Clear Rx Frame Drop Interrupt */
//        if( IntStatus & MAC_INT_RX_FRAME_DROPPED )
//        {
//            MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_DROPPED );
//            IntStatus &= ~(MAC_INT_RX_FRAME_DROPPED);
//        }
//
////DEBUGMSG(LAN_DBG_LVL,"LANISR : LANReceive = %x\n", nRxDone );
//        LANReceive( pEthNetInfo );
//        //hbp_eth_RecvHandler (dev);
//    }
//
//    // handle transmitter interrupt
//    if( IntStatus & MAC_INT_TX_DONE )
//    {
////DEBUGMSG(LAN_DBG_LVL,"LANISR : MAC_INT_TX_DONE\n" );
//        // clear TX_DONE INT
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_TX_DONE );
//
//        if( pEthNetInfo->TxStatus == TRUE )
//        {
//            // Read number of BADDR remaining in TX_BADDR
//            MAC_REG_READ( pEthNetInfo, TX_BUFLVL, Temp);
//            TxFifoLevel = (Temp & 0x3F00) >> 8;
//
//            /* nTxFrame = nTxDone + TxFifoLevel + 1                 */
//            /* nTxDoneN is calculated approximately with safety.    */
//            /* Actual Tx Done Number may be ntxDone or nTxDone+1    */
//            pEthNetInfo->nTxDone = pEthNetInfo->nTxFrame - TxFifoLevel - 1;
//            /* Though nTxFrame can be overflowed, it is not a problem               */
//            /* becuase nTxFrame is type [unsigned int] and 0xFFFFFFFF + 1 = 0       */
//            /*  0 - 1 = 0xFFFFFFFF, 0 - 2 = 0xFFFFFFFE  , ...                       */
//
//            /* Free Tx done buffer to be used again */
//            while( pEthNetInfo->nTxFree != pEthNetInfo->nTxDone )
//            {
//                pBufferProcessed = pEthNetInfo->pNextTxDone;
//
//                MAC_REG_READ( pEthNetInfo, TX_DONE_BADDR, Temp );
//
//                //pEthNetInfo->Stats.Tx_packets++;
//                //dev_kfree_skb_irq (bufferProcessed->skb);
//                pBufferProcessed->Status = _B_FREE_;
//                pEthNetInfo->pNextTxDone = pBufferProcessed->pNext;
//                pEthNetInfo->nTxFree++;
//            }
//
//            for (i = 0; i < MAC_TX_BADDR_REG_FIFO_DEPTH - TxFifoLevel; i++)
//            {
//                pBufferToProcess = pEthNetInfo->pNextDeviceTx;
//            
//                // check next buffer is filled
//                // there exists buffer(s) to send
//                if( pBufferToProcess->Status == _B_FILLED_)
//                {
//                    // link new buffer to the device -> it will awake the device
//                    MAC_REG_WRITE( pEthNetInfo, TX_BADDR, (UINT)(pBufferToProcess->Buf) );
//                    MAC_REG_WRITE( pEthNetInfo, TX_LENGTH, pBufferToProcess->FrameLength );
//
//                    pEthNetInfo->nTxFrame++;
//            
//                    // update next processing pointer (advance the pointer)
//                    pEthNetInfo->pNextDeviceTx = pBufferToProcess->pNext;
//                }
//                else        /* next buffer is free buffer */
//                {
//                    if (i == 0 && TxFifoLevel == 0 )
//                    {
//                        pEthNetInfo->TxStatus = FALSE;    // transmitter idle
//                    }
//            
//                    break;
//                }
//            }
//        }
//        else
//        {
//            /* NONE */
//            /* Ignore Tx Done INT after changed to priv->txStatus = FALSE */
//        }
//    }
//
//    // handle dropped 
//    if( IntStatus & MAC_INT_RX_FRAME_DROPPED )
//    {
////DEBUGMSG(LAN_DBG_LVL,"LANISR : MAC_INT_RX_FRAME_DROPPED\n" );
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_RX_FRAME_DROPPED );
//        pBufferToProcess = pEthNetInfo->pNextDeviceRx;
//
//        if( pBufferToProcess->Status == _B_FREE_ )  // there exists free buffer(s)
//        {
//            /* link new buffer to the device -> it will awake the device */
//            MAC_REG_WRITE( pEthNetInfo, RX_BADDR, (UINT32)pBufferToProcess->Buf );
//
//            /* update next processing pointer (advance the pointer) */
//            pEthNetInfo->pNextDeviceRx = pBufferToProcess->pNext;
//
//            //new_skb = dev_alloc_skb (FRAME_MAX_AL);
//            //
//              //        if (new_skb != NULL)
//            //                {
//              //                new_skb->data += PAYLOAD_ALIGN_OFFSET;
//              //                new_skb->tail += PAYLOAD_ALIGN_OFFSET;
//              //                bufferToProcess->skb = new_skb;
//              //                bufferToProcess->pCluster = new_skb->data - PAYLOAD_ALIGN_OFFSET;
//            //
//              //                /* link new buffer to the device -> it will awake the device */
//              //                MAC_REG_WRITE (priv, RX_BADDR, (UINT) bufferToProcess->pCluster);
//            //
//              //                /* update next processing pointer (advance the pointer) */
//              //                priv->nextDeviceRx = bufferToProcess->next;
//            //                }
//              //        else
//            //                {
//              //                /* Do Nothing */
//            //                }
//        }
//        else
//        {
//            // Disable MAC_INT_RX_FRAME_DROPPED
//            MAC_REG_READ( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, Temp);
//            MAC_REG_WRITE( pEthNetInfo, HBP_MAC_INT_ENABLE_REG, Temp & ~(MAC_INT_RX_FRAME_DROPPED) );
//        }
//
//        //hbp_eth_RecvHandler (dev);
//        LANReceive( pEthNetInfo );
//    }
//
//      // MII
//      if( IntStatus & MAC_INT_MII_PHY )
//    {
////DEBUGMSG(LAN_DBG_LVL,"LANISR : MAC_INT_MII_PHY\n" );
//        if( pEthNetInfo->PhyType == PHY_KS8737 )
//        {
//            UINT32  PhyIntStatus;
//
//            // read PHY INT status register
//            // PHY interrupt will be automatically cleared by reading register
//            PhyIntStatus = PhyRegRead( pEthNetInfo, PHY_KS8737_INT_ENABLE_STATUS);
//
//            if( PhyIntStatus & (PHY_KS8737_INT_LINK_UP|PHY_KS8737_INT_LINK_DOWN) )
//            {
//                if ( /* check if eth attached ? */ 1)
//                {
//                    if(PhyLinkStatusRead( pEthNetInfo ) )
//                    {
//                        /* Link UP */
//                        //netif_carrier_on (dev);
//                    }    
//                    else
//                    {
//                        /* Link Down */
//                        //netif_carrier_off (dev);
//                    }
//                }
//            }
//        }
//    
//        // clear MII_PHY Int bit of MAC
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, MAC_INT_MII_PHY );
//    }
//
//    // fatal error
//    if( IntStatus & MAC_INT_FATAL_ERROR )
//    {
////DEBUGMSG(LAN_DBG_LVL,"LANISR : MAC_INT_FATAL_ERROR\n" );
//        MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus & MAC_INT_FATAL_ERROR );
//    }
//
//INT_RETURN:
////    MAC_REG_READ( pEthNetInfo, INT_SRC, IntStatus);
////DEBUGMSG(LAN_DBG_LVL,"LANISR : IntStatus = %x\n", IntStatus );
////    MAC_REG_WRITE( pEthNetInfo, INT_SRC, IntStatus);
//
//    return;
//    //SYS_INT_ACK (priv->ivec);
//    //spin_unlock (&priv->lock);
//}


STATIC BOOL
thLANDeviceDriver( UINT32 Arg )
{

    JB_RECVMESSAGE      Message;
    HANDLE              hPort;
    //HANDLE              Handle = NULL;
    PENET_INFO          pEthNetInfo;
    NET_PDEV			pNetDev = (NET_PDEV)Arg;

    hPort = PortGet( );    
    PortSetType( hPort, PORTTYPE_MESSAGE|PORTTYPE_INTERRUPT|PORTTYPE_SIGNAL );

    LAN_DeviceConfig.Name[3] = (UINT32)LAN_DeviceConfig.Minor + '1';
    LAN_DeviceConfig.Channel = ChannelGet( hPort );

    DeviceRegister( &LAN_DeviceConfig );
DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver pDeviceConfig->Channel = %X\n",LAN_DeviceConfig.Channel);

    pEthNetInfo = &EthNetInfo;
    //pEthNetInfo->iDev = iDev;

DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver LANInit pNetDev = %x\n",pNetDev );
DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver LANInit &LAN_NetDev = %x\n",&LAN_NetDev );


DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver LANInit pEthNetInfo = %X\n",pEthNetInfo );
    LANInit( pEthNetInfo, pNetDev);

DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver InterruptRegister pEthNetInfo->nINTVirt = %d\n",pEthNetInfo->nINTVirt );
    InterruptRegister( pEthNetInfo->nINTVirt, hPort);

    // open routine
DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver LANOpen pEthNetInfo = %X\n",pEthNetInfo );
    LANOpen( pEthNetInfo );

    pEthNetInfo->Handle = ChannelOpenHandleGetFrom( LAN_DeviceConfig.Channel );
    DEBUGMSG(LAN_DBG_LVL,"thDeviceDriverThreadHead Handle = %X\n",pEthNetInfo->Handle);

    pNetDev->pDrvName = &LAN_DeviceConfig.Name[0];
    pNetDev->Handle = pEthNetInfo->Handle;

    pEthNetInfo->iNetIf = NwRegisterDev( pNetDev );

    InterruptEnable(pEthNetInfo->nINTVirt);

    while(1)
    {
        MessageReceive( hPort, &Message);
//DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MessageReceive\n" );

        switch(Message.Type)
        {
        case MESSAGE_TYPE_SIGNAL:
        {
            if( Message.Param & SIG_RESERVED_ALARM)
            {
            }

            if( Message.Param & SIG_NETWORK_OUTPUT)
            {
                BOOL            bRet;

//DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MESSAGE_TYPE_SIGNAL %d\n", pEthNetInfo->iDev);
NEXT_PACKET:
                bRet = LANSend( pEthNetInfo );
//DEBUGMSG(LAN_DBG_LVL,"thLANDeviceDriver : MESSAGE_TYPE_SIGNAL bRet = %d\n", bRet);
                if( bRet )
                    goto NEXT_PACKET;

            }

            if( Message.Param & SIG_NETWORK_SET)
            {
                NET_DEV_SET   NetDevSet;
                
                UINT32  Resp = -ERR_NETWORK_OPNOTSUPP;
DEBUGMSG( LAN_DBG_LVL,"%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
                // get set data from network server
                if( NwDevGetSet( pNetDev->iNetIf, &NetDevSet, sizeof(NET_DEV_SET) ) )
                {
                    // response the set data 
                    NwDevRespSet( pEthNetInfo->iNetIf, Resp );
                }
            }

#if 0 // case of wireless network 
            if( Message.Param & SIG_NETWORK_SET)
            {
                NET_DEV_SET   NetDevSet;
                
                UINT32  Resp = -EOPNOTSUPP;
                // get the network set
                struct ifreq *p_ifreq;
                UINT32          Cmd;

                // get set data from network server
                if( NwDevGetSet( pNetDev->iNetIf, &NetDevSet, sizeof(NET_DEV_SET) ) )
                {
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET Cmd = %x\n", __FUNCTION__, __LINE__, NetDevSet.Cmd);
//kprintf("%s (%d): SIG_NETWORK_SET Cmd = %x\n", __FUNCTION__, __LINE__, NetDevSet.Cmd);
                    /* check the wireless standard control */
                    if( NetDevSet.Cmd >= SOCK_CTRL_WLAN_FIRST && 
                        NetDevSet.Cmd <= SOCK_CTRL_WLAN_LAST)
                    {

#if 1
                        Resp = 
                        wireless_process_ioctl(net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd);
#else
                        iw_handler  handler;
                        struct iw_handler_def *pwh;
                        pwh = net_dev->wireless_handlers;
                        
                        if( pwh->standard == NULL ||
                            IW_IOCTL_IDX(NetDevSet.Cmd) >= pwh->num_standard)
                            goto SET_DONE;

                        PNWDEVREQ pNwDevReq = (PNWDEVREQ)NetDevSet.pSet;
                    	struct iw_request_info info = { .cmd = NetDevSet.Cmd, .flags = 0 };

                        handler = pwh->standard[IW_IOCTL_IDX(NetDevSet.Cmd)];

                        Resp = handler(net_dev, &info, &pNwDevReq->NwDevReqU, NULL/*extra*/);
#endif
                    }
                    else
                    if( NetDevSet.Cmd >= SOCK_CTRL_WLAN_FIRST && 
                        NetDevSet.Cmd <= SOCK_CTRL_WLAN_LAST)
                    {
#if 1
                        Resp = 
                        wireless_process_ioctl(net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd);
#else
                        iw_handler  handler;
                        struct iw_handler_def *pwh;
                        pwh = net_dev->wireless_handlers;
                        
                        if( pwh->private == NULL ||
                            IW_IOCTL_IDX(NetDevSet.Cmd) >= pwh->num_private)
                            goto SET_DONE;

                        PNWDEVREQ pNwDevReq = (PNWDEVREQ)NetDevSet.pSet;
                    	struct iw_request_info info = { .cmd = NetDevSet.Cmd, .flags = 0 };

                        handler = pwh->standard[IW_IOCTL_IDX(NetDevSet.Cmd)];

                        Resp = handler(net_dev, &info, &pNwDevReq->NwDevReqU, NULL/*extra*/);
#endif
                    }
                    else
                    {
                        Resp = 
                        net_dev->do_ioctl( net_dev, (struct ifreq *)NetDevSet.pSet, NetDevSet.Cmd );
                    }
SET_DONE:
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
//kprintf("%s (%d): SIG_NETWORK_SET Resp  %x\n", __FUNCTION__, __LINE__, Resp);
                    // response the set data 
                    NwDevRespSet( pNetDev->iNetIf, Resp );
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET ++\n", __FUNCTION__, __LINE__);
                }
DEBUGMSG( DBG_LVL_WLAN_SET,"%s (%d): SIG_NETWORK_SET ++\n", __FUNCTION__, __LINE__);
            }
#endif

            continue;
        }

        case MESSAGE_TYPE_INTERRUPT:
        {
DEBUGMSG(LAN_DBG_LVL,"%s(%d) : MESSAGE_TYPE_INTERRUPT %d\n",__FUNCTION__,__LINE__, pEthNetInfo->nINTVirt);
            LANISR( pEthNetInfo );
            InterruptDone( pEthNetInfo->nINTVirt );
DEBUGMSG(LAN_DBG_LVL,"%s(%d): MESSAGE_TYPE_INTERRUPT END\n",__FUNCTION__,__LINE__);
            continue;
        }

        case MESSAGE_TYPE_DEVICE_READ:
        {
            Message.Type = 0;
            break;
        }

        case MESSAGE_TYPE_DEVICE_WRITE:
        {
            Message.Type = 0;
            break;
        }

        case MESSAGE_TYPE_DEVICE_CONTROL:
        {
            Message.Type = FALSE;
            break;
        }

        case MESSAGE_TYPE_DEVICE_OPEN:
        {
            //DEBUGMSG(1,"MESSAGE_TYPE_DEVICE_OPEN %X\n", Channel);
            //if( Handle != NULL)
            //{
            //    ThreadSetLastError( Message.hSender, ERR_BUSY );
            //    Message.Type = NULL;
            //    break;
            //}
            //
            //LANOpen();
            //
            //Handle = ChannelOpenHandleGetFrom( Channel );
            //DEBUGMSG(1,"thDeviceDriverThreadHead %X\n",Handle);
            //
            ////InterruptEnable(SYSINTR_LAN);
            //
            //Message.Type = Handle;

            Message.Type = NULL;
            break;
        }

        case MESSAGE_TYPE_DEVICE_CLOSE:
        {
            //DEBUGMSG(1,"MESSAGE_TYPE_DEVICE_CLOSE %X\n", Handle);
            //DEBUGMSG(1,"MESSAGE_TYPE_DEVICE_CLOSE %X\n", Message.Handle);

            if( Message.Handle != pEthNetInfo->Handle )
            {
                ThreadSetLastError( Message.hSender, ERR_INVALID );
                Message.Type = FALSE;
                break;
            }

            LANClose( pEthNetInfo );
            InterruptDisable( pEthNetInfo->nINTVirt );

            pEthNetInfo->Handle = NULL;
            Message.Type = TRUE;
            break;
        }

        default:
            break;
        }

        MessageReply(&Message);

    }

    return TRUE;    
}




#include <moduleloader.h>

PUBLIC HANDLE
DllMain( HANDLE hDll, UINT32 Reason, PVOID Reserved )
{
    HANDLE hThread = NULL;
    switch( Reason )
    {
    case DLL_PROCESS_ATTACH:
        hThread = ThreadCreate (&GMACB_LANDev_Thread);
    	break;
    
    case DLL_PROCESS_DETACH:
    	break;
    }

    return hThread;
}




