//    KSZ9021

#include <types.h>
#include <Time.h>
#include <panic.h>
#include "phy.h"

//-----------------------------------------------------------------------------
///         Definitions
//-----------------------------------------------------------------------------
//IEEE defined Registers
#define GMII_BMCR        0   // Basic Mode Control Register
#define GMII_BMSR        1   // Basic Mode Status Register
#define GMII_PHYID1      2   // PHY Idendifier Register 1
#define GMII_PHYID2      3   // PHY Idendifier Register 2
#define GMII_ANAR        4   // Auto_Negotiation Advertisement Register
#define GMII_ANLPAR      5   // Auto_negotiation Link Partner Ability Register
#define GMII_ANER        6   // Auto-negotiation Expansion Register
#define GMII_ANNPR       7   // Auto-negotiation Next Page Register
#define GMII_ANLPNPAR    8   // Auto_negotiation Link Partner Next Page Ability Register
#define GMII_1000BTCR    9   // 1000Base-T Control
#define GMII_1000BTSR   10   // 1000Base-T Status
#define GMII_ERCR       11   // Extend Register - Control Register
#define GMII_ERDWR      12   // Extend Register - Data Write Register
#define GMII_ERDRR      13   // Extend Register - Data Read Register
//14    reserved
#define GMII_EMSR       15   // Extend MII Status Register

//Vender Specific Register
//16    reserved
#define GMII_RLLMR      17   // Remote Loopback, LED Mode Register
#define GMII_LMDCDR     18   // LinkND - Cable Diagnostic Register
#define GMII_DPPSR      19   // Digital PMA/PCS Status
//20    reserved
#define GMII_RXERCR     21   // RXER Counter Register
//22-26 reserved
#define GMII_ICSR       27   // Interrupt Control/Status Register
#define GMII_DDC1R      28   // Digital Debug Control 1 Register
//29-30 reserved
#define GMII_PHYCR      31   // PHY Control Register

//Extend Registers
#define GMII_CCR        256  // Common Control Register
#define GMII_SSR        257  // Strap Status Register
#define GMII_OMSOR      258  // Operation Mode Strap Override Register
#define GMII_OMSSR      259  // Operation Mode Strap Status Register
#define GMII_RCCPSR     260  // RGMII Clock and Control Pad Skew Register
#define GMII_RRDPSR     261  // RGMII RX Data Pad Skew Register
#define GMII_ATR        263  // Analog Test Register



// Basic Mode Control Register (BMCR)
// Bit definitions: MII_BMCR
#define GMII_RESET             (1 << 15) // 1= Software Reset; 0=Normal Operation
#define GMII_LOOPBACK          (1 << 14) // 1=loopback Enabled; 0=Normal Operation
#define GMII_SPEED_SELECT_LSB  (1 << 13) // 1,0=1000Mbps 0,1=100Mbps; 0,0=10Mbps
#define GMII_AUTONEG           (1 << 12) // Auto-negotiation Enable
#define GMII_POWER_DOWN        (1 << 11) // 1=Power down 0=Normal operation
#define GMII_ISOLATE           (1 << 10) // 1 = Isolates 0 = Normal operation
#define GMII_RESTART_AUTONEG   (1 << 9)  // 1 = Restart auto-negotiation 0 = Normal operation
#define GMII_DUPLEX_MODE       (1 << 8)  // 1 = Full duplex operation 0 = Normal operation
//      Reserved                7        // Read as 0, ignore on write
#define GMII_SPEED_SELECT_MSB  (1 << 6)  // 
//      Reserved                5 to 0   // Read as 0, ignore on write

// Basic Mode Status Register (BMSR)
// Bit definitions: MII_BMSR
#define GMII_100BASE_T4        (1 << 15) // 100BASE-T4 Capable
#define GMII_100BASE_TX_FD     (1 << 14) // 100BASE-TX Full Duplex Capable
#define GMII_100BASE_T4_HD     (1 << 13) // 100BASE-TX Half Duplex Capable
#define GMII_10BASE_T_FD       (1 << 12) // 10BASE-T Full Duplex Capable
#define GMII_10BASE_T_HD       (1 << 11) // 10BASE-T Half Duplex Capable
//      Reserved                10 to 9  // Read as 0, ignore on write
#define GMII_EXTEND_STATUS     (1 << 8)  // 1 = Extend Status Information In Reg 15
//      Reserved                7
#define GMII_MF_PREAMB_SUPPR   (1 << 6)  // MII Frame Preamble Suppression
#define GMII_AUTONEG_COMP      (1 << 5)  // Auto-negotiation Complete
#define GMII_REMOTE_FAULT      (1 << 4)  // Remote Fault
#define GMII_AUTONEG_ABILITY   (1 << 3)  // Auto Configuration Ability
#define GMII_LINK_STATUS       (1 << 2)  // Link Status
#define GMII_JABBER_DETECT     (1 << 1)  // Jabber Detect
#define GMII_EXTEND_CAPAB      (1 << 0)  // Extended Capability

// PHY ID Identifier Register
// definitions: MII_PHYID1
#define GMII_LSB_MASK           0x3F
#define GMII_OUI_MSB            0x0022
#define GMII_OUI_LSB            0x05

// Auto-negotiation Advertisement Register (ANAR)
// Auto-negotiation Link Partner Ability Register (ANLPAR)
// Bit definitions: MII_ANAR, MII_ANLPAR
#define GMII_NP               (1 << 15) // Next page Indication
//      Reserved               7
#define GMII_RF               (1 << 13) // Remote Fault
//      Reserved               12       // Write as 0, ignore on read
#define GMII_PAUSE_MASK       (3 << 11) // 0,0 = No Pause 1,0 = Asymmetric Pause(link partner)
                                        // 0,1 = Symmetric Pause 1,1 = Symmetric&Asymmetric Pause(local device)   
#define GMII_T4               (1 << 9)  // 100BASE-T4 Support
#define GMII_TX_FDX           (1 << 8)  // 100BASE-TX Full Duplex Support
#define GMII_TX_HDX           (1 << 7)  // 100BASE-TX Support
#define GMII_10_FDX           (1 << 6)  // 10BASE-T Full Duplex Support
#define GMII_10_HDX           (1 << 5)  // 10BASE-T Support
//      Selector                 4 to 0   // Protocol Selection Bits
#define GMII_AN_IEEE_802_3      0x0001

// Auto-negotiation Expansion Register (ANER)
// Bit definitions: MII_ANER
//      Reserved                15 to 5  // Read as 0, ignore on write
#define GMII_PDF              (1 << 4) // Local Device Parallel Detection Fault
#define GMII_LP_NP_ABLE       (1 << 3) // Link Partner Next Page Able
#define GMII_NP_ABLE          (1 << 2) // Local Device Next Page Able
#define GMII_PAGE_RX          (1 << 1) // New Page Received
#define GMII_LP_AN_ABLE       (1 << 0) // Link Partner Auto-negotiation Able

// GMII_1000BTCR
#define GMII_1000BaseT_HALF_DUPLEX       (1 << 8)  
#define GMII_1000BaseT_FULL_DUPLEX       (1 << 9)  
#define GMII_MARSTER_SLAVE_ENABLE        (1 << 12)  
#define GMII_MARSTER_SLAVE_CONFIG        (1 << 11)  
#define GMII_PORT_TYPE                   (1 << 10)  

// GMII_1000BTSR  
#define GMII_LINKP_1000BaseT_HALF_DUPLEX (1 << 10)  
#define GMII_LINKP_1000BaseT_FULL_DUPLEX (1 << 11)  



STATIC BOOL 
PHY_Reset(MAC_PHY_POP pMacPhyOp, BYTE PhyAddr)
{
    UINT32  bmcr = GMII_RESET;
    UINT32  timeout = 10;

    DEBUGMSG(1,"RESET PHY!!!\r\n");
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_BMCR, GMII_RESET);

    do
    {
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_BMCR, &bmcr);
        timeout--;
    } while ((bmcr & GMII_RESET) && timeout);

    if (!timeout)
        return FALSE;

    return TRUE;
}

STATIC BOOL 
PHY_StartAutoNegociation(MAC_PHY_POP pMacPhyOp, BYTE PhyAddr)
{
    BOOL        bRet = TRUE;
    UINT32      Value;
    UINT32      PhyAnar;
    UINT32      gbaseTC;

    Value = GMII_RCCPSR | 0x8000;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERCR, Value);
    //value = 0xF0F4;
    Value = 0xF2F4;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERDWR, Value);
    Value = GMII_RRDPSR | 0x8000;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERCR, Value);
    //value = 0x3028;
    Value = 0x2222;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERDWR, Value);
    
    Value = 0xFF00;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ICSR, Value);
 
    /* Set the Auto_negotiation Advertisement Register, MII advertising for Next page
       100BaseTxFD and HD, 10BaseTFD and HD, IEEE 802.3 */
    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                            GMII_ANAR, &PhyAnar);
//    if (bRet == 0) 
//    {
//        goto AutoNegotiateExit;
//    }
    PhyAnar = GMII_TX_FDX | GMII_TX_HDX |
              GMII_10_FDX | GMII_10_HDX | GMII_AN_IEEE_802_3;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ANAR, PhyAnar);
//    if (bRet == 0) 
//    {
//        goto AutoNegotiateExit;
//    }
//


    /* Read & modify 1000Base-T control register  */
    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_1000BTCR, &gbaseTC);

    gbaseTC |= GMII_1000BaseT_HALF_DUPLEX |GMII_1000BaseT_FULL_DUPLEX;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_1000BTCR, gbaseTC);
    
    /* Read & modify control register */
    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_BMCR, &Value);

    /* Restart Auto_negotiation */
    Value |=  GMII_RESTART_AUTONEG;
    pMacPhyOp->WritePhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_BMCR, Value);


//AutoNegotiateExit:
    return bRet;
}



STATIC BOOL 
PHY_WaitForLink(MAC_PHY_POP pMacPhyOp, BYTE PhyAddr, UINT32 TimeOut)
{
    BOOL    bRet = FALSE;
    UINT32  Value;

    TimeOut = TimeOut / 100;


    /* Check AutoNegotiate complete */
    do
    {
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                            GMII_DPPSR, &Value);

        /* Done successfully */
        if( Value & ((1<<1)|(1<<2)) ) 
        {
            kprintf("Link complete\n\r");
            bRet = TRUE;
            break;
        }

        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                            GMII_BMSR, &Value);
        if (Value & GMII_LINK_STATUS )
        {
            kprintf("Link complete\n\r");
            bRet = TRUE;
            break;
        }

        /* Timeout check */
        Sleep(100);
    }while( TimeOut-- > 0 );

    return bRet;
}

STATIC BOOL 
PHY_WaitForAutonegociationComplete( MAC_PHY_POP pMacPhyOp, BYTE PhyAddr, UINT32 TimeOut )
{
    BOOL    bRet = FALSE;
    UINT32  Value;

    TimeOut = TimeOut / 100;


    /* Check AutoNegotiate complete */
    do
    {
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                            GMII_BMSR, &Value);

        /* Done successfully */
        if (Value & GMII_AUTONEG_COMP) 
        {
            kprintf("AutoNegotiate complete\n\r");
            bRet = TRUE;
            break;
        }

        /* Timeout check */
        Sleep(100);
    }while( TimeOut-- > 0 );

    return bRet;
}


STATIC BOOL 
PHY_GetID(MAC_PHY_POP pMacPhyOp, BYTE PhyAddr, PUINT32 pID)
{
    UINT32      ID, Value, PhyAnar;

    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_PHYID1, &Value);
//kprintf("Value = 0x%X\n\r", Value);
    
    pMacPhyOp->ReadPhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_PHYID2, &PhyAnar);
//kprintf("PhyAnar = 0x%X\n\r", PhyAnar);

    if( ( Value == GMII_OUI_MSB )
     && ( ((PhyAnar>>10)&GMII_LSB_MASK) == GMII_OUI_LSB ) )
    {
//        kprintf("Vendor Number Model = 0x%X\n\r", ((PhyAnar>>4)&0x3F));
//        kprintf("Model Revision Number = 0x%X\n\r", (PhyAnar&0x7));
        return TRUE;
    }
    else 
    {
        kprintf("Problem OUI value\n\r");
        return FALSE;
    }
}

STATIC BOOL 
PHY_CheckID( MAC_PHY_POP pMacPhyOp, BYTE PhyAddr)
{
    UINT32      ID;

    BOOL        bRet;
    bRet = PHY_GetID( pMacPhyOp,PhyAddr,&ID);

    return bRet;
}


STATIC BOOL 
PHY_SetConfiguration( MAC_PHY_POP pMacPhyOp, BYTE PhyAddr, PHY_PINFO pPhyInfo )
{    
    UINT32  CommCtrl;
    UINT32  ControlReg;

    // Check parameters !
    if( pPhyInfo == NULL )
    {
        return FALSE;
    }


	/* Timing configuration for KSZ9021RN PHY */
	pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, 11, 260 | 0x8000);
	pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, 12, 0xF2F4);
	pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, 11, 261 | 0x8000);
	pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, 12, 0x2222);

   
    //pMacPhyOp->ReadPhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_CCR, &CommCtrl);

    pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERCR, GMII_CCR);
    pMacPhyOp->ReadPhy( pMacPhyOp->pDeviceLocation, PhyAddr, 
                        GMII_ERDRR, &CommCtrl);

        
    if( (pPhyInfo->bRMII == FALSE)  && (CommCtrl & (1<<8)) )
    {
        DEBUGMSG(1,"PHY_SetConfiguration : MII mode not available\r\n");
        return FALSE;
    }

    if( (pPhyInfo->bRMII == TRUE)  && (!(CommCtrl & (1<<8))))
    {
        DEBUGMSG(1,"PHY_SetConfiguration : RMII mode not available\r\n");
        return FALSE;
    }

    if( pPhyInfo->bAutoNegociation )
    {
        BOOL    bRet;

        bRet = PHY_StartAutoNegociation( pMacPhyOp, PhyAddr );

        //pMacPhyOp->ReadPhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_BMCR, &ControlReg);
        //ControlReg |= (1<<14);
        //pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_BMCR, ControlReg);    
        //pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_DDC1R, (1<<0) );    

//int i;
//for( i = 0; i < 31 ; i ++ )
//{
//    UINT32 Val;
//pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation,PhyAddr, i, &Val );
//kprintf("%d = 0x%x\n", i, Val);
//}

        return bRet;
    }

    // Check parameters
    if( (pPhyInfo->Speed != SPEED_1000) && 
        (pPhyInfo->Speed != SPEED_100) && 
        (pPhyInfo->Speed != SPEED_10) )
    {
        return FALSE;
    }

    ControlReg = 0;
    if( pPhyInfo->Speed  == SPEED_1000)
    {
        ControlReg |= (0<<13)|(1<<6);
    }
    else
    if( pPhyInfo->Speed  == SPEED_100 )
    {
        ControlReg |= (1<<13)|(0<<6);
    }
    else
    //if( pPhyInfo->Speed  == SPEED_10 )
    {
        ControlReg |= (0<<13)|(0<<6);
    }

    if( pPhyInfo->bFullDuplex )
    {
        ControlReg |= (1<<8);
    }

    pMacPhyOp->WritePhy( pMacPhyOp->pDeviceLocation, PhyAddr, GMII_BMCR, ControlReg);    
   
    return FALSE;
}



STATIC BOOL
PHY_GetConfiguration( MAC_PHY_POP pMacPhyOp, BYTE PhyAddr, PHY_PINFO pPhyInfo)
{
    BOOL    bRet = FALSE;
    UINT32  Value;


    UINT32  gbaseTS;
    UINT32  gbaseTC;
    UINT32  PhyAnalpar;
    UINT32  PhyAnar;

    if( pPhyInfo == NULL )
    {
        return bRet;
    }

    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_BMSR, &Value);
    if (Value & GMII_AUTONEG_COMP) 
    {
        pPhyInfo->bAutoNegociation = TRUE;    

        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                                GMII_ANAR, &PhyAnar);

        /* Read & modify 1000Base-T control register  */
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                                GMII_1000BTCR, &gbaseTC);
        gbaseTC |= GMII_1000BaseT_HALF_DUPLEX |GMII_1000BaseT_FULL_DUPLEX;

        /*Set local link mode */
        while(1)
        {
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_1000BTSR, &gbaseTS);

        /* Setup the EMAC link speed */
        if ((gbaseTS & GMII_LINKP_1000BaseT_FULL_DUPLEX) &&
          (gbaseTC & GMII_1000BaseT_FULL_DUPLEX))
        {
            /* set RGMII for 1000BaseTX and Full Duplex */
            pPhyInfo->bFullDuplex = TRUE;//GMAC_DUPLEX_FULL;
            pPhyInfo->Speed = SPEED_1000;
            break;
        }
        else
        if ((gbaseTS & GMII_LINKP_1000BaseT_HALF_DUPLEX) &&
          (gbaseTC & GMII_1000BaseT_HALF_DUPLEX))
        {
            /* set RGMII for 1000BaseT and Half Duplex*/
            pPhyInfo->bFullDuplex = FALSE;//GMAC_DUPLEX_HALF;
            pPhyInfo->Speed = SPEED_1000;
            break;
        }

        /* Get the AutoNeg Link partner base page */
        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_ANLPAR, &PhyAnalpar);

        /* Setup the EMAC link speed */
        if ((PhyAnar & PhyAnalpar) & GMII_TX_FDX) 
        {
            /* set RGMII for 100BaseTX and Full Duplex */
            pPhyInfo->bFullDuplex = TRUE;//GMAC_DUPLEX_FULL;
            pPhyInfo->Speed = SPEED_100;
            break;
        }
        else
        if ((PhyAnar & PhyAnalpar) & GMII_10_FDX) 
        {
            /* set RGMII for 10BaseT and Full Duplex */
            pPhyInfo->bFullDuplex = TRUE;//GMAC_DUPLEX_FULL;
            pPhyInfo->Speed = SPEED_10;
            break;
        }
        else
        if ((PhyAnar & PhyAnalpar) & GMII_TX_HDX) 
        {
            /* set RGMII for 100BaseTX and half Duplex */
            pPhyInfo->bFullDuplex = FALSE;//GMAC_DUPLEX_HALF;
            pPhyInfo->Speed = SPEED_100;
            break;
        }
        else
        if ((PhyAnar & PhyAnalpar) & GMII_10_HDX) 
        {
            /* set RGMII for 10BaseT and half Duplex */
            pPhyInfo->bFullDuplex = FALSE;//GMAC_DUPLEX_HALF;
            pPhyInfo->Speed = SPEED_10;
            break;
        }

        }

bRet = TRUE;
    } 
    else 
    {
        UINT32  Value;
        pPhyInfo->bAutoNegociation = FALSE;

        pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, GMII_1000BTSR, &gbaseTS);

        /* Setup the EMAC link speed */
        if ((gbaseTS & GMII_LINKP_1000BaseT_FULL_DUPLEX) &&
          (gbaseTC & GMII_1000BaseT_FULL_DUPLEX))
        {
            /* set RGMII for 1000BaseTX and Full Duplex */
            pPhyInfo->bFullDuplex = TRUE;//GMAC_DUPLEX_FULL;
            pPhyInfo->Speed = SPEED_1000;
            //break;
        }
        else
        if ((gbaseTS & GMII_LINKP_1000BaseT_HALF_DUPLEX) &&
          (gbaseTC & GMII_1000BaseT_HALF_DUPLEX))
        {
            /* set RGMII for 1000BaseT and Half Duplex*/
            pPhyInfo->bFullDuplex = FALSE;//GMAC_DUPLEX_HALF;
            pPhyInfo->Speed = SPEED_1000;
            //break;
        }
        else
        {
            pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation, PhyAddr, 
                                GMII_PHYCR, &Value);

#define MIIM_KSZ9021_PHYCTL_1000	(1 << 6)
#define MIIM_KSZ9021_PHYCTL_100		(1 << 5)
#define MIIM_KSZ9021_PHYCTL_10		(1 << 4)
#define MIIM_KSZ9021_PHYCTL_DUPLEX	(1 << 3)


            if (Value & MIIM_KSZ9021_PHYCTL_1000 )
            {
                pPhyInfo->Speed = SPEED_1000;
            }
            else
            if (Value & MIIM_KSZ9021_PHYCTL_100 )
            {    
                pPhyInfo->Speed = SPEED_100;
            }
            else
            if (Value & MIIM_KSZ9021_PHYCTL_10 )
            {
                pPhyInfo->Speed = SPEED_10;
            }

            if (Value & MIIM_KSZ9021_PHYCTL_DUPLEX )
                pPhyInfo->bFullDuplex = TRUE;
            else
                pPhyInfo->bFullDuplex = FALSE;

        }
    }

//    pMacPhyOp->ReadPhy(pMacPhyOp->pDeviceLocation,PhyAddr,DM9161A_DAVICOM_SPEC_CONF_REG,&DavicomSpecifiedConfReg);
//        
//    if( DavicomSpecifiedConfReg & DM9161A_DAVICOM_SPEC_CONF_RMII_ENABLED )
//    {
        pPhyInfo->bRMII = TRUE;        
//    }
//    else
//    {
//        pPhyInfo->bRMII = FALSE;
//    }

    return bRet;
}


STATIC  BYTE PhyName[] = "KSZ9201";


PUBLIC PHY_INTERFACE KSZ9201_PhyInterface =
{
    PhyName,
    PHY_Reset,
    PHY_SetConfiguration,
    PHY_GetConfiguration,    
    PHY_StartAutoNegociation,
    PHY_WaitForLink,
    PHY_WaitForAutonegociationComplete,
    PHY_GetID,    
    PHY_CheckID,
};



